from __future__ import annotations

from ctypes import sizeof, c_uint32 as uint32
from typing import Optional, TYPE_CHECKING

from ..qt_compat import QByteArray

from ..exception import (
    AccountAuthKeyNotFound,
    Expects,
    ExpectStreamStatus,
    OpenTeleException,
    TDataInvalidMagic,
)
from ..utils import BaseObject
from .configs import dbi
from . import shared as td
from .map_data import MapData

if TYPE_CHECKING:
    from .account import Account


class StorageAccount(BaseObject):  # nocov
    def __init__(self, owner: Account, basePath: str, keyFile: str) -> None:
        self.__owner = owner
        self.__keyFile = keyFile
        self.__dataNameKey = td.Storage.ComputeDataNameKey(self.__keyFile)
        self.__baseGlobalPath = td.Storage.GetAbsolutePath(basePath)
        self.__basePath = td.Storage.PathJoin(
            self.__baseGlobalPath, td.Storage.ToFilePart(self.__dataNameKey)
        )
        self.__localKey = None
        self.__mapData = MapData(self.basePath)
        self.__config = td.MTP.Config(td.MTP.Environment.Production)

    @property
    def owner(self) -> Account:
        return self.__owner

    @property
    def localKey(self) -> Optional[td.AuthKey]:
        return self.__localKey

    @localKey.setter
    def localKey(self, value):
        self.__localKey = value

    @property
    def keyFile(self) -> str:
        return self.__keyFile

    @keyFile.setter
    def keyFile(self, value):
        self.__keyFile = value
        self.__dataNameKey = td.Storage.ComputeDataNameKey(self.__keyFile)
        self.baseGlobalPath = self.baseGlobalPath

    @property
    def baseGlobalPath(self) -> str:
        return self.__baseGlobalPath

    @baseGlobalPath.setter
    def baseGlobalPath(self, basePath):
        self.__baseGlobalPath = td.Storage.GetAbsolutePath(basePath)
        self.__basePath = td.Storage.PathJoin(
            self.__baseGlobalPath, td.Storage.ToFilePart(self.__dataNameKey)
        )

    @property
    def basePath(self) -> str:
        return self.__basePath

    @property
    def config(self) -> td.MTP.Config:
        return self.__config

    @property
    def mapData(self) -> MapData:
        return self.__mapData

    def start(self, localKey: td.AuthKey) -> td.MTP.Config:
        self.__localKey = localKey
        self._readMapWith(localKey)
        return (
            self.__config
            if self.owner.owner.kPerformanceMode
            else self._readMtpConfig()
        )

    def _readMtpData(self):
        mtp = td.Storage.ReadEncryptedFile(
            td.Storage.ToFilePart(self.__dataNameKey),
            self.__baseGlobalPath,
            self.localKey,
        )  # type: ignore

        blockId = mtp.stream.readInt32()
        Expects(blockId == 75, TDataInvalidMagic("Not supported file version"))

        serialized = QByteArray()
        mtp.stream >> serialized
        self.owner._setMtpAuthorization(serialized)

    def _readMtpConfig(self) -> td.MTP.Config:
        Expects(
            self.localKey is not None,
            AccountAuthKeyNotFound("The localKey has not been initialized yet"),
        )

        try:
            file = td.Storage.ReadEncryptedFile("config", self.basePath, self.localKey)  # type: ignore
            serialized = QByteArray()
            file.stream >> serialized
            ExpectStreamStatus(file.stream, "Could not stream data from MtpConfig")
            self.__config = td.MTP.Config.FromSerialized(serialized)
            return self.__config
        except OpenTeleException:
            pass

        return td.MTP.Config(td.MTP.Environment.Production)

    def _readMapWith(
        self, localKey: td.AuthKey, legacyPasscode: QByteArray = QByteArray()
    ):
        try:
            self.__mapData.read(localKey, legacyPasscode)
        except OpenTeleException:
            return False

        self._readMtpData()

    def _writeMtpConfig(self, basePath: str) -> None:
        Expects(self.localKey is not None, "localKey not found")
        Expects(basePath is not None and basePath != "", "basePath can't be empty")

        serialized = self.owner.MtpConfig.Serialize()
        size = td.Serialize.bytearraySize(serialized)
        file = td.Storage.FileWriteDescriptor("config", basePath)
        data = td.Storage.EncryptedDescriptor(size)
        data.stream << serialized
        file.writeEncrypted(data, self.localKey)  # type: ignore
        file.finish()

    def _writeMap(self, basePath: str) -> None:
        Expects(self.localKey is not None, "localKey not found")
        Expects(basePath is not None and basePath != "", "basePath can't be empty")

        map = td.Storage.FileWriteDescriptor("map", basePath)
        map.writeData(QByteArray())
        map.writeData(QByteArray())

        mapDataEncrypted = self.mapData.prepareToWrite()
        map.writeEncrypted(mapDataEncrypted, self.localKey)  # type: ignore
        map.finish()

    def _writeMtpData(self, baseGlobalPath: str, dataNameKey: int) -> None:
        Expects(self.localKey is not None, "localKey not found")
        Expects(
            baseGlobalPath is not None and baseGlobalPath != "",
            "baseGlobalPath can't be empty",
        )

        serialized = self.owner.serializeMtpAuthorization()
        size = sizeof(uint32) + td.Serialize.bytearraySize(serialized)
        mtp = td.Storage.FileWriteDescriptor(
            td.Storage.ToFilePart(dataNameKey), baseGlobalPath
        )
        data = td.Storage.EncryptedDescriptor(size)
        data.stream.writeInt32(dbi.MtpAuthorization)
        data.stream << serialized
        mtp.writeEncrypted(data, self.localKey)  # type: ignore
        mtp.finish()

    def _writeData(self, baseGlobalPath: str, keyFile: str = None) -> None:
        Expects(
            baseGlobalPath is not None and baseGlobalPath != "",
            "baseGlobalPath can't be empty",
        )

        if keyFile is not None and self.keyFile != keyFile:
            self.__keyFile = keyFile
            dataNameKey = td.Storage.ComputeDataNameKey(self.__keyFile)
        else:
            dataNameKey = self.__dataNameKey

        basePath = td.Storage.PathJoin(
            baseGlobalPath, td.Storage.ToFilePart(dataNameKey)
        )
        self._writeMap(basePath)

        if not self.owner.owner.kPerformanceMode:
            self._writeMtpConfig(basePath)

        self._writeMtpData(baseGlobalPath, dataNameKey)



from __future__ import annotations
import inspect
import types
import typing
from .qt_compat import QDataStream

__all__ = [
    "OpenTeleException",
    "TFileNotFound",
    "TDataInvalidMagic",
    "TDataInvalidCheckSum",
    "TDataBadDecryptKey",
    "TDataWrongPasscode",
    "TDataBadEncryptedDataSize",
    "TDataBadDecryptedDataSize",
    "TDataBadConfigData",
    "QDataStreamFailed",
    "AccountAuthKeyNotFound",
    "TDataReadMapDataFailed",
    "TDataReadMapDataIncorrectPasscode",
    "TDataAuthKeyNotFound",
    "MaxAccountLimit",
    "TDesktopUnauthorized",
    "TelethonUnauthorized",
    "TDataSaveFailed",
    "TDesktopNotLoaded",
    "TDesktopHasNoAccount",
    "TDAccountNotLoaded",
    "NoPasswordProvided",
    "PasswordIncorrect",
    "LoginFlagInvalid",
    "NoInstanceMatched",
    "SessionFileNotFound",
    "SessionFileInvalid",
    "Expects",
    "ExpectStreamStatus",
]


class OpenTeleException(BaseException):  # nocov
    """Base exception of the library."""

    def __init__(self, message: str = None, stack_index: int = 1) -> None:
        super().__init__(message if (message is not None) else "")

        self.message = message
        self.desc = self.__class__.__name__

        self.frame = inspect.currentframe()
        for i in range(stack_index):
            self.frame = self.frame.f_back

        self._caller_class = (
            self.frame.f_locals["self"].__class__
            if "self" in self.frame.f_locals
            else None
        )
        self._caller_method = self.frame.f_code.co_name

        if self._caller_method != "<module>":
            args, _, _, locals = inspect.getargvalues(self.frame)
            parameters = {arg: locals[arg] for arg in args}
            self._caller_method_params = "".join(
                f"{i}={parameters[i]}, " for i in parameters
            )[:-2]
        else:
            self._caller_method = "__main__"
            self._caller_method_params = ""

        if self.desc == "OpenTeleException":
            self.desc = "Unexpected Exception"

    def __str__(self):
        reason = self.desc.__str__()

        if self.message is not None:
            reason += f": {self.message}"

        reason += " [ Called by "
        if self._caller_class is not None:
            parent_list = []
            base = self._caller_class
            while hasattr(base, "__name__"):
                parent_list.append(base.__name__)
                base = base.__base__

            parent_list.reverse()
            reason += "".join(f"{i}." for i in parent_list[1:])
            reason += self._caller_method + "() ]"
        else:
            reason += f"{self._caller_method}() ]"
        return reason


class TFileNotFound(OpenTeleException):
    pass


class TDataInvalidMagic(OpenTeleException):
    pass


class TDataInvalidCheckSum(OpenTeleException):
    pass


class TDataBadDecryptKey(OpenTeleException):
    pass


class TDataWrongPasscode(OpenTeleException):
    pass


class TDataBadEncryptedDataSize(OpenTeleException):
    pass


class TDataBadDecryptedDataSize(OpenTeleException):
    pass


class TDataBadConfigData(OpenTeleException):
    pass


class QDataStreamFailed(OpenTeleException):
    pass


class AccountAuthKeyNotFound(OpenTeleException):
    pass


class TDataReadMapDataFailed(OpenTeleException):
    pass


class TDataReadMapDataIncorrectPasscode(OpenTeleException):
    pass


class TDataAuthKeyNotFound(OpenTeleException):
    pass


class MaxAccountLimit(OpenTeleException):
    pass


class TDesktopUnauthorized(OpenTeleException):
    pass


class TelethonUnauthorized(OpenTeleException):
    pass


class TDataSaveFailed(OpenTeleException):
    pass


class TDesktopNotLoaded(OpenTeleException):
    pass


class TDesktopHasNoAccount(OpenTeleException):
    pass


class TDAccountNotLoaded(OpenTeleException):
    pass


class NoPasswordProvided(OpenTeleException):
    pass


class PasswordIncorrect(OpenTeleException):
    pass


class LoginFlagInvalid(OpenTeleException):
    pass


class NoInstanceMatched(OpenTeleException):
    pass


class SessionFileNotFound(OpenTeleException):
    pass


class SessionFileInvalid(OpenTeleException):
    pass


@typing.overload
def Expects(
    condition: bool,
    message: str = None,
    done: typing.Callable[[], None] = None,
    fail: typing.Callable[[OpenTeleException], None] = None,
    silent: bool = False,
    stack_index: int = 1,
) -> bool: ...


@typing.overload
def Expects(  # noqa: F811
    condition: bool,
    exception: OpenTeleException = None,
    done: typing.Callable[[], None] = None,
    fail: typing.Callable[[OpenTeleException], None] = None,
    silent: bool = False,
    stack_index: int = 1,
) -> bool: ...


def Expects(  # noqa: F811
    condition: bool,
    exception: typing.Union[OpenTeleException, str] = None,
    done: typing.Callable[[], None] = None,
    fail: typing.Callable[[OpenTeleException], None] = None,
    silent: bool = False,
    stack_index: int = 1,
) -> bool:  # nocov
    if condition:
        if done is not None:
            done()
        return condition

    if isinstance(exception, str):
        exception = OpenTeleException(exception, 2)
    elif exception is not None and not isinstance(exception, OpenTeleException):
        raise OpenTeleException("No instance of Expects() match the arguments given", 2)

    if exception is None:
        exception = OpenTeleException("Unexpected error", 2)

    if silent:
        if fail is not None:
            fail(exception)
        return condition

    stack = inspect.stack()
    frame = stack[stack_index].frame
    tb = types.TracebackType(None, frame, frame.f_lasti, frame.f_lineno)  # type: ignore
    exception = exception.with_traceback(tb)

    if fail is not None:
        fail(exception)

    raise exception


def ExpectStreamStatus(stream: QDataStream, message: str = "Could not stream data"):
    Expects(
        stream.status() == QDataStream.Status.Ok,
        stack_index=2,
        exception=QDataStreamFailed(message, stack_index=2),
    )



"""Comprehensive offline test suite for opentele2.

Tests all functionality: API, devices, fingerprint, consistency,
tdata loading, session files, JSON consistency, and conversion.
"""

import os
import sys
import json
import sqlite3
import pathlib

# ---------------------------------------------------------------------------
# Setup: path
# ---------------------------------------------------------------------------
base_dir = pathlib.Path(__file__).parent.parent.absolute().__str__()
sys.path.insert(0, base_dir)

# ---------------------------------------------------------------------------
# Imports (after path setup)
# ---------------------------------------------------------------------------
import pytest  # noqa: E402

from src.td import TDesktop  # noqa: E402
from src.td.account import Account  # noqa: E402
from src.tl.telethon import TelegramClient  # noqa: E402
from src.api import API, APIData, UseCurrentSession, CreateNewSession, LoginFlag  # noqa: E402
from src.devices import (  # noqa: E402
    AndroidDevice,
    IOSDevice,
    macOSDevice,
    WindowsDevice,
    LinuxDevice,
    WebBrowserDevice,
    DeviceInfo,
)
from src.fingerprint import (  # noqa: E402
    LAYER,
    PLATFORM_VERSIONS,
    StrictMode,
    FingerprintConfig,
    DEFAULT_CONFIG,
    TransportRecommendation,
    validate_init_connection_params,
    get_recommended_layer,
    get_platform_versions,
    generate_msg_id_offset,
    is_valid_msg_id,
)
from src.consistency import (  # noqa: E402
    CheckResult,
    ConsistencyReport,
)

# ---------------------------------------------------------------------------
# Constants
# ---------------------------------------------------------------------------
TESTS_DIR = pathlib.Path(__file__).parent
TDATAS_DIR = TESTS_DIR / "tdatas"
SESSIONS_DIR = TESTS_DIR / "sessions"

ACCOUNT_IDS = ["215340804", "215342020", "215342289", "215342316", "215342360"]


def _load_json(account_id: str) -> dict:
    path = SESSIONS_DIR / f"{account_id}.json"
    with open(path, "r", encoding="utf-8") as f:
        return json.load(f)


def _read_session_db(account_id: str) -> dict:
    """Read dc_id and auth_key from a Telethon .session SQLite file."""
    path = SESSIONS_DIR / f"{account_id}.session"
    conn = sqlite3.connect(str(path))
    try:
        row = conn.execute(
            "SELECT dc_id, server_address, port, auth_key FROM sessions"
        ).fetchone()
        if row is None:
            return {}
        return {
            "dc_id": row[0],
            "server_address": row[1],
            "port": row[2],
            "auth_key": row[3],
        }
    finally:
        conn.close()


# ===================================================================
# 1. API MODULE TESTS
# ===================================================================


class TestAPIClasses:
    """Tests for the API class hierarchy."""

    def test_all_api_classes_exist(self):
        apis = [
            API.TelegramDesktop,
            API.TelegramAndroid,
            API.TelegramAndroidX,
            API.TelegramIOS,
            API.TelegramMacOS,
            API.TelegramWeb_Z,
            API.TelegramWeb_A,
            API.TelegramWeb_K,
            API.Webogram,
        ]
        for api in apis:
            assert issubclass(api, APIData), f"{api.__name__} must extend APIData"

    def test_api_ids_are_official(self):
        """Each API class should use a known official api_id."""
        official_ids = {2040, 6, 21724, 10840, 2834, 2496}
        apis = [
            API.TelegramDesktop,
            API.TelegramAndroid,
            API.TelegramAndroidX,
            API.TelegramIOS,
            API.TelegramMacOS,
            API.TelegramWeb_Z,
            API.TelegramWeb_A,
            API.TelegramWeb_K,
            API.Webogram,
        ]
        for api in apis:
            assert api.api_id in official_ids, (
                f"{api.__name__}.api_id={api.api_id} not in official IDs"
            )

    def test_api_hashes_non_empty(self):
        apis = [
            API.TelegramDesktop,
            API.TelegramAndroid,
            API.TelegramAndroidX,
            API.TelegramIOS,
            API.TelegramMacOS,
            API.TelegramWeb_Z,
            API.TelegramWeb_A,
            API.TelegramWeb_K,
            API.Webogram,
        ]
        for api in apis:
            assert api.api_hash and len(api.api_hash) == 32, (
                f"{api.__name__}.api_hash invalid"
            )

    def test_desktop_attributes(self):
        d = API.TelegramDesktop
        assert d.api_id == 2040
        assert d.api_hash == "b18441a1ff607e10a989891a5462e627"
        assert d.lang_pack == "tdesktop"
        assert d.lang_code == "en"

    def test_android_attributes(self):
        a = API.TelegramAndroid
        assert a.api_id == 6
        assert a.lang_pack == "android"

    def test_ios_attributes(self):
        ios = API.TelegramIOS
        assert ios.api_id == 10840
        assert ios.lang_pack == "ios"

    def test_macos_attributes(self):
        m = API.TelegramMacOS
        assert m.api_id == 2834
        assert m.lang_pack == "macos"

    def test_web_z_attributes(self):
        w = API.TelegramWeb_Z
        assert w.api_id == 2496
        assert "Chrome" in w.device_model

    def test_api_instantiation(self):
        """Instantiating an API class should copy the class defaults."""
        inst = API.TelegramDesktop()
        assert inst.api_id == API.TelegramDesktop.api_id
        assert inst.api_hash == API.TelegramDesktop.api_hash
        assert inst.lang_pack == API.TelegramDesktop.lang_pack

    def test_api_copy(self):
        """copy() should produce an equivalent but distinct instance."""
        original = API.TelegramAndroid()
        copied = original.copy()
        assert copied.api_id == original.api_id
        assert copied.api_hash == original.api_hash
        assert copied.device_model == original.device_model
        assert copied.system_version == original.system_version

    def test_api_custom_instantiation(self):
        custom = APIData(api_id=12345, api_hash="a" * 32)
        assert custom.api_id == 12345
        assert custom.api_hash == "a" * 32

    def test_api_str_representation(self):
        s = str(API.TelegramDesktop)
        assert "TelegramDesktop" in s


class TestAPIGenerate:
    """Tests for API.Generate() methods."""

    def test_desktop_generate_deterministic(self):
        a = API.TelegramDesktop.Generate("windows", "seed123")
        b = API.TelegramDesktop.Generate("windows", "seed123")
        assert a.device_model == b.device_model
        assert a.system_version == b.system_version

    def test_desktop_generate_different_systems(self):
        win = API.TelegramDesktop.Generate("windows", "same_seed")
        linux = API.TelegramDesktop.Generate("linux", "same_seed")
        assert win.system_version != linux.system_version

    def test_desktop_generate_random_varies(self):
        """Without unique_id, successive calls should produce different results."""
        seen = set()
        for _ in range(10):
            g = API.TelegramDesktop.Generate()
            seen.add((g.device_model, g.system_version))
        assert len(seen) > 1, "Random generation should produce variety"

    def test_android_generate(self):
        g = API.TelegramAndroid.Generate("seed")
        assert g.api_id == API.TelegramAndroid.api_id
        assert g.device_model is not None
        assert g.system_version is not None

    def test_ios_generate(self):
        g = API.TelegramIOS.Generate("seed")
        assert g.api_id == API.TelegramIOS.api_id
        assert "iPhone" in g.device_model

    def test_macos_generate(self):
        g = API.TelegramMacOS.Generate("seed")
        assert g.api_id == API.TelegramMacOS.api_id

    def test_android_x_generate(self):
        g = API.TelegramAndroidX.Generate("seed")
        assert g.api_id == API.TelegramAndroidX.api_id


class TestLoginFlag:
    def test_use_current_session(self):
        assert issubclass(UseCurrentSession, LoginFlag)

    def test_create_new_session(self):
        assert issubclass(CreateNewSession, LoginFlag)


# ===================================================================
# 2. DEVICE GENERATION TESTS
# ===================================================================


class TestAndroidDevice:
    def test_random_device_returns_device_info(self):
        d = AndroidDevice.RandomDevice("test_seed")
        assert isinstance(d, DeviceInfo)
        assert d.model
        assert d.version

    def test_deterministic(self):
        a = AndroidDevice.RandomDevice("same")
        b = AndroidDevice.RandomDevice("same")
        assert a.model == b.model
        assert a.version == b.version

    def test_random_varies(self):
        results = set()
        for _ in range(20):
            d = AndroidDevice.RandomDevice()
            results.add(d.model)
        assert len(results) > 1

    def test_sdk_version_format(self):
        d = AndroidDevice.RandomDevice("check_format")
        assert d.version.startswith("SDK ")
        sdk_num = int(d.version.replace("SDK ", ""))
        assert 21 <= sdk_num <= 40


class TestIOSDevice:
    def test_random_device(self):
        d = IOSDevice.RandomDevice("test")
        assert isinstance(d, DeviceInfo)
        assert "iPhone" in d.model

    def test_deterministic(self):
        a = IOSDevice.RandomDevice("same_seed")
        b = IOSDevice.RandomDevice("same_seed")
        assert a.model == b.model
        assert a.version == b.version

    def test_version_format(self):
        d = IOSDevice.RandomDevice("check")
        parts = d.version.split(".")
        assert 2 <= len(parts) <= 3
        assert all(p.isdigit() for p in parts)

    def test_device_list_populated(self):
        IOSDevice.__gen__()
        assert len(IOSDevice.deviceList) > 100


class TestMacOSDevice:
    def test_random_device(self):
        d = macOSDevice.RandomDevice("test")
        assert isinstance(d, DeviceInfo)

    def test_deterministic(self):
        a = macOSDevice.RandomDevice("seed")
        b = macOSDevice.RandomDevice("seed")
        assert a.model == b.model


class TestWindowsDevice:
    def test_random_device(self):
        d = WindowsDevice.RandomDevice("test")
        assert isinstance(d, DeviceInfo)

    def test_version_is_windows(self):
        d = WindowsDevice.RandomDevice("win_test")
        assert "Windows" in d.version or "10" in d.version or "11" in d.version


class TestLinuxDevice:
    def test_random_device(self):
        d = LinuxDevice.RandomDevice("test")
        assert isinstance(d, DeviceInfo)

    def test_version_contains_linux(self):
        d = LinuxDevice.RandomDevice("linux_test")
        assert "Linux" in d.version


class TestWebBrowserDevice:
    def test_random_device_z_variant(self):
        d = WebBrowserDevice.RandomDevice("test", variant="z")
        assert isinstance(d, DeviceInfo)
        assert "Mozilla" in d.model or "Chrome" in d.model or len(d.model) > 20

    def test_random_device_k_variant(self):
        d = WebBrowserDevice.RandomDevice("test", variant="k")
        assert isinstance(d, DeviceInfo)

    def test_z_vs_k_system_version_differs(self):
        """Z variant returns OS name, K variant returns navigator.platform."""
        dz = WebBrowserDevice.RandomDevice("same_seed", variant="z")
        dk = WebBrowserDevice.RandomDevice("same_seed", variant="k")
        assert dz.model == dk.model  # same UA
        # system_version may differ (e.g., "Windows" vs "Win32")

    def test_deterministic(self):
        a = WebBrowserDevice.RandomDevice("seed", variant="z")
        b = WebBrowserDevice.RandomDevice("seed", variant="z")
        assert a.model == b.model
        assert a.version == b.version


# ===================================================================
# 3. FINGERPRINT MODULE TESTS
# ===================================================================


class TestFingerprintValidation:
    def test_valid_desktop_params(self):
        issues = validate_init_connection_params(
            api_id=2040,
            device_model="Desktop",
            system_version="Windows 11",
            app_version="5.12.3 x64",
            system_lang_code="en-US",
            lang_pack="tdesktop",
            lang_code="en",
        )
        assert issues == []

    def test_valid_android_params(self):
        issues = validate_init_connection_params(
            api_id=6,
            device_model="Samsung Galaxy S24 Ultra",
            system_version="SDK 35",
            app_version="12.3.0",
            system_lang_code="en-US",
            lang_pack="android",
            lang_code="en",
        )
        assert issues == []

    def test_invalid_lang_pack(self):
        issues = validate_init_connection_params(
            api_id=2040,
            device_model="Desktop",
            system_version="Windows 11",
            app_version="5.12.3",
            system_lang_code="en",
            lang_pack="invalid_pack",
            lang_code="en",
        )
        assert any("lang_pack" in i for i in issues)

    def test_empty_device_model(self):
        issues = validate_init_connection_params(
            api_id=2040,
            device_model="",
            system_version="Windows 11",
            app_version="5.12.3",
            system_lang_code="en",
            lang_pack="tdesktop",
            lang_code="en",
        )
        assert any("device_model" in i for i in issues)

    def test_short_lang_code(self):
        issues = validate_init_connection_params(
            api_id=2040,
            device_model="Desktop",
            system_version="Windows 11",
            app_version="5.12.3",
            system_lang_code="en",
            lang_pack="tdesktop",
            lang_code="x",
        )
        assert any("lang_code" in i for i in issues)

    def test_strict_mode_api_id_mismatch(self):
        issues = validate_init_connection_params(
            api_id=9999,
            device_model="Desktop",
            system_version="Windows 11",
            app_version="5.12.3",
            system_lang_code="en-US",
            lang_pack="tdesktop",
            lang_code="en",
            strict=True,
        )
        assert any("api_id" in i for i in issues)

    def test_strict_mode_all_correct(self):
        issues = validate_init_connection_params(
            api_id=2040,
            device_model="Desktop",
            system_version="Windows 11",
            app_version="5.12.3 x64",
            system_lang_code="en-US",
            lang_pack="tdesktop",
            lang_code="en",
            strict=True,
        )
        assert issues == []

    def test_mobile_system_lang_needs_region(self):
        issues = validate_init_connection_params(
            api_id=6,
            device_model="Samsung Galaxy S24",
            system_version="SDK 35",
            app_version="12.3.0",
            system_lang_code="en",  # missing region
            lang_pack="android",
            lang_code="en",
        )
        assert any("region" in i for i in issues)

    def test_web_empty_lang_pack_valid(self):
        issues = validate_init_connection_params(
            api_id=2496,
            device_model="Mozilla/5.0 ...",
            system_version="Windows",
            app_version="5.0.0 Z",
            system_lang_code="en-US",
            lang_pack="",
            lang_code="en",
        )
        assert issues == []


class TestPlatformVersions:
    def test_singleton(self):
        pv = get_platform_versions()
        assert pv is PLATFORM_VERSIONS

    def test_android_version(self):
        assert PLATFORM_VERSIONS.android_app_version
        parts = PLATFORM_VERSIONS.android_app_version.split(".")
        assert len(parts) >= 2

    def test_ios_version(self):
        assert PLATFORM_VERSIONS.ios_app_version

    def test_desktop_version(self):
        assert PLATFORM_VERSIONS.desktop_app_version

    def test_chrome_version(self):
        assert PLATFORM_VERSIONS.chrome_version
        major = int(PLATFORM_VERSIONS.chrome_version.split(".")[0])
        assert major >= 100

    def test_user_agent_contains_chrome(self):
        assert "Chrome" in PLATFORM_VERSIONS.user_agent

    def test_sdk_range(self):
        lo, hi = PLATFORM_VERSIONS.android_sdk_range
        assert lo < hi
        assert lo >= 21


class TestFingerprintConfig:
    def test_default_config(self):
        assert DEFAULT_CONFIG.strict_mode == StrictMode.WARN
        assert DEFAULT_CONFIG.auto_validate is True
        assert DEFAULT_CONFIG.preferred_transport == "obfuscated"

    def test_strict_mode_values(self):
        assert StrictMode.OFF.value == "off"
        assert StrictMode.WARN.value == "warn"
        assert StrictMode.STRICT.value == "strict"

    def test_config_validate_params_warn(self):
        """Config in WARN mode should not raise, just warn."""
        config = FingerprintConfig(strict_mode=StrictMode.WARN)
        import warnings as w

        with w.catch_warnings(record=True) as caught:
            w.simplefilter("always")
            config.validate_params(
                api_id=9999,
                device_model="",
                system_version="",
                app_version="",
                system_lang_code="en",
                lang_pack="bad",
                lang_code="en",
            )
        assert len(caught) > 0

    def test_config_validate_params_strict_raises(self):
        config = FingerprintConfig(strict_mode=StrictMode.STRICT)
        with pytest.raises(ValueError, match="fingerprint"):
            config.validate_params(
                api_id=9999,
                device_model="",
                system_version="",
                app_version="",
                system_lang_code="en",
                lang_pack="bad",
                lang_code="en",
            )

    def test_config_validate_off_does_nothing(self):
        config = FingerprintConfig(strict_mode=StrictMode.OFF, auto_validate=False)
        # Should not raise or warn
        config.validate_params(
            api_id=0,
            device_model="",
            system_version="",
            app_version="",
            system_lang_code="",
            lang_pack="bad",
            lang_code="",
        )

    def test_effective_layer(self):
        config = FingerprintConfig()
        layer = config.get_effective_layer()
        assert isinstance(layer, int)
        assert layer > 100

    def test_layer_override(self):
        config = FingerprintConfig(layer_override=999)
        assert config.get_effective_layer() == 999


class TestMsgIdHelpers:
    def test_generate_offset(self):
        offset = generate_msg_id_offset()
        assert 0 <= offset <= 0xFFFF

    def test_valid_client_msg_id(self):
        import time

        now = int(time.time())
        msg_id = (now << 32) | 1  # odd = client
        assert is_valid_msg_id(msg_id, from_client=True)

    def test_invalid_zero_msg_id(self):
        assert not is_valid_msg_id(0)

    def test_invalid_parity(self):
        import time

        now = int(time.time())
        even_id = (now << 32) | 2
        assert not is_valid_msg_id(even_id, from_client=True)  # client must be odd

    def test_server_msg_id(self):
        import time

        now = int(time.time())
        even_id = (now << 32) | 2
        assert is_valid_msg_id(even_id, from_client=False)


class TestTransportRecommendation:
    def test_get_connection_class(self):
        cls = TransportRecommendation.get_connection_class()
        assert cls is not None

    def test_available_transports(self):
        transports = TransportRecommendation.get_available_transports()
        assert isinstance(transports, dict)
        assert len(transports) > 0

    def test_default_is_obfuscated(self):
        cls = TransportRecommendation.get_connection_class("tdesktop")
        assert "Obfuscated" in cls.__name__ or "Full" in cls.__name__


class TestLayer:
    def test_layer_is_int(self):
        assert isinstance(LAYER, int)
        assert LAYER > 100

    def test_recommended_layer(self):
        layer = get_recommended_layer()
        assert isinstance(layer, int)
        assert layer > 100


# ===================================================================
# 4. CONSISTENCY MODULE TESTS
# ===================================================================


class TestConsistencyDataclasses:
    def test_check_result_passed(self):
        cr = CheckResult(name="test", passed=True, detail="OK")
        assert cr.passed
        assert cr.name == "test"
        assert cr.detail == "OK"

    def test_check_result_failed(self):
        cr = CheckResult(name="test", passed=False, detail="bad")
        assert not cr.passed

    def test_report_all_passed(self):
        r = ConsistencyReport(
            checks=[
                CheckResult(name="a", passed=True, detail="ok"),
                CheckResult(name="b", passed=True, detail="ok"),
            ]
        )
        assert r.all_passed

    def test_report_not_all_passed(self):
        r = ConsistencyReport(
            checks=[
                CheckResult(name="a", passed=True, detail="ok"),
                CheckResult(name="b", passed=False, detail="bad"),
            ]
        )
        assert not r.all_passed

    def test_report_summary(self):
        r = ConsistencyReport(
            checks=[
                CheckResult(name="a", passed=True, detail="ok"),
                CheckResult(name="b", passed=False, detail="fail"),
            ]
        )
        s = r.summary
        assert "1/2" in s
        assert "[OK] a" in s
        assert "[FAIL] b" in s

    def test_report_str(self):
        r = ConsistencyReport(checks=[CheckResult(name="x", passed=True, detail="d")])
        assert str(r) == r.summary

    def test_empty_report_passes(self):
        r = ConsistencyReport()
        assert r.all_passed  # no checks = all pass


# ===================================================================
# 5. TDATA LOADING TESTS
# ===================================================================


class TestTDataLoading:
    """Test loading TDesktop tdata folders for all 5 test accounts."""

    @pytest.fixture(params=ACCOUNT_IDS)
    def account_id(self, request):
        return request.param

    @pytest.fixture
    def tdata_path(self, account_id):
        return str(TDATAS_DIR / account_id / "tdata")

    @pytest.fixture
    def account_json(self, account_id):
        return _load_json(account_id)

    def test_tdata_loads_successfully(self, tdata_path):
        tdesk = TDesktop(tdata_path)
        assert tdesk.isLoaded()

    def test_tdata_has_accounts(self, tdata_path):
        tdesk = TDesktop(tdata_path)
        assert tdesk.accountsCount > 0

    def test_tdata_main_account_exists(self, tdata_path):
        tdesk = TDesktop(tdata_path)
        assert tdesk.mainAccount is not None

    def test_tdata_account_is_loaded(self, tdata_path):
        tdesk = TDesktop(tdata_path)
        assert tdesk.mainAccount.isLoaded()

    def test_tdata_account_has_auth_key(self, tdata_path):
        tdesk = TDesktop(tdata_path)
        acct = tdesk.mainAccount
        assert acct.authKey is not None
        assert len(acct.authKey.key) == 256

    def test_tdata_account_has_user_id(self, tdata_path, account_id):
        tdesk = TDesktop(tdata_path)
        acct = tdesk.mainAccount
        assert acct.UserId > 0

    def test_tdata_account_has_valid_dc_id(self, tdata_path):
        tdesk = TDesktop(tdata_path)
        acct = tdesk.mainAccount
        assert 1 <= acct.MainDcId <= 5

    def test_tdata_user_id_matches_json(self, tdata_path, account_id):
        """The user ID from tdata should be a valid positive integer."""
        tdesk = TDesktop(tdata_path)
        acct = tdesk.mainAccount
        assert acct.UserId > 0, f"Invalid UserId for account {account_id}"

    def test_tdata_keyfile(self, tdata_path):
        tdesk = TDesktop(tdata_path)
        assert tdesk.keyFile == "data"

    def test_tdata_serialization_roundtrip(self, tdata_path):
        """Serialize and deserialize MTP authorization should preserve data."""
        tdesk = TDesktop(tdata_path)
        acct = tdesk.mainAccount
        serialized = acct.serializeMtpAuthorization()
        assert len(serialized) > 0

        # The serialized data should start with the wide IDs tag
        from src.qt_compat import QDataStream

        stream = QDataStream(serialized)
        stream.setVersion(QDataStream.Version.Qt_5_1)
        tag = stream.readInt64()
        assert tag == Account.kWideIdsTag


# ===================================================================
# 6. SESSION FILE TESTS
# ===================================================================


class TestSessionFiles:
    """Test that .session SQLite files are valid and readable."""

    @pytest.fixture(params=ACCOUNT_IDS)
    def account_id(self, request):
        return request.param

    def test_session_file_exists(self, account_id):
        path = SESSIONS_DIR / f"{account_id}.session"
        assert path.exists()

    def test_session_file_readable(self, account_id):
        data = _read_session_db(account_id)
        assert data, f"Could not read session {account_id}"

    def test_session_has_dc_id(self, account_id):
        data = _read_session_db(account_id)
        assert 1 <= data["dc_id"] <= 5

    def test_session_has_auth_key(self, account_id):
        data = _read_session_db(account_id)
        assert data["auth_key"] is not None
        assert len(data["auth_key"]) == 256

    def test_session_has_server_address(self, account_id):
        data = _read_session_db(account_id)
        assert data["server_address"]

    def test_session_port_is_443(self, account_id):
        data = _read_session_db(account_id)
        assert data["port"] == 443


# ===================================================================
# 7. JSON CONSISTENCY TESTS
# ===================================================================


class TestJSONConsistency:
    """Cross-reference JSON metadata with session/tdata data."""

    @pytest.fixture(params=ACCOUNT_IDS)
    def account_id(self, request):
        return request.param

    @pytest.fixture
    def json_data(self, account_id):
        return _load_json(account_id)

    def test_json_has_required_fields(self, json_data):
        required = [
            "app_id",
            "app_hash",
            "device",
            "sdk",
            "app_version",
            "lang_pack",
            "lang_code",
            "session_file",
        ]
        for field in required:
            assert field in json_data, f"Missing field: {field}"

    def test_json_app_id_is_official(self, json_data):
        official_ids = {2040, 6, 21724, 10840, 2834, 2496}
        assert json_data["app_id"] in official_ids

    def test_json_app_hash_length(self, json_data):
        assert len(json_data["app_hash"]) == 32

    def test_json_session_file_matches(self, account_id, json_data):
        assert json_data["session_file"] == account_id

    def test_json_lang_code_valid(self, json_data):
        assert len(json_data["lang_code"]) >= 2

    def test_json_device_not_empty(self, json_data):
        assert json_data["device"]

    def test_json_sdk_not_empty(self, json_data):
        assert json_data["sdk"]

    def test_tdata_session_authkey_match(self, account_id):
        """Auth key from tdata must match auth key from .session file."""
        tdata_path = str(TDATAS_DIR / account_id / "tdata")
        tdesk = TDesktop(tdata_path)
        tdata_key = tdesk.mainAccount.authKey.key

        session_data = _read_session_db(account_id)
        session_key = session_data["auth_key"]

        assert tdata_key == session_key, f"Auth key mismatch for account {account_id}"

    def test_tdata_session_dc_id_match(self, account_id):
        """DC ID from tdata must match DC ID from .session file."""
        tdata_path = str(TDATAS_DIR / account_id / "tdata")
        tdesk = TDesktop(tdata_path)
        tdata_dc = int(tdesk.mainAccount.MainDcId)

        session_data = _read_session_db(account_id)
        session_dc = session_data["dc_id"]

        assert tdata_dc == session_dc, (
            f"DC ID mismatch for {account_id}: tdata={tdata_dc}, session={session_dc}"
        )

    def test_json_api_matches_known_api_class(self, json_data):
        """The app_id and app_hash in JSON should match a known API class."""
        api_map = {
            2040: API.TelegramDesktop,
            6: API.TelegramAndroid,
            21724: API.TelegramAndroidX,
            10840: API.TelegramIOS,
            2834: API.TelegramMacOS,
            2496: API.TelegramWeb_Z,  # Web Z/A/K all share 2496
        }
        app_id = json_data["app_id"]
        assert app_id in api_map
        api_cls = api_map[app_id]
        assert json_data["app_hash"] == api_cls.api_hash

    def test_json_fingerprint_validates(self, json_data):
        """The JSON metadata should pass basic fingerprint validation."""
        issues = validate_init_connection_params(
            api_id=json_data["app_id"],
            device_model=json_data["device"],
            system_version=json_data["sdk"],
            app_version=json_data["app_version"],
            system_lang_code=json_data.get("system_lang_code", "en"),
            lang_pack=json_data.get("lang_pack", ""),
            lang_code=json_data["lang_code"],
        )
        # Only check for critical issues (empty fields)
        critical = [i for i in issues if "empty" in i.lower()]
        assert len(critical) == 0, f"Critical fingerprint issues: {critical}"


# ===================================================================
# 8. CONVERSION TESTS (offline, no network)
# ===================================================================


class TestConversion:
    """Test TDesktop -> TelegramClient conversion (UseCurrentSession, offline)."""

    @pytest.fixture(params=ACCOUNT_IDS)
    def account_id(self, request):
        return request.param

    @pytest.mark.asyncio
    async def test_tdata_to_telethon_creates_client(self, account_id, tmp_path):
        """Convert tdata to TelegramClient using UseCurrentSession."""
        tdata_path = str(TDATAS_DIR / account_id / "tdata")
        tdesk = TDesktop(tdata_path)
        assert tdesk.isLoaded()

        session_path = str(tmp_path / f"test_{account_id}")
        client = await tdesk.ToTelethon(
            session=session_path,
            flag=UseCurrentSession,
            api=API.TelegramDesktop,
        )
        assert client is not None
        assert client.session is not None
        assert client.session.auth_key is not None
        assert len(client.session.auth_key.key) == 256

    @pytest.mark.asyncio
    async def test_conversion_preserves_auth_key(self, account_id, tmp_path):
        """Auth key should be preserved through TDesktop -> TelegramClient."""
        tdata_path = str(TDATAS_DIR / account_id / "tdata")
        tdesk = TDesktop(tdata_path)

        original_key = tdesk.mainAccount.authKey.key

        session_path = str(tmp_path / f"test_conv_{account_id}")
        client = await tdesk.ToTelethon(
            session=session_path,
            flag=UseCurrentSession,
            api=API.TelegramDesktop,
        )

        assert client.session.auth_key.key == original_key

    @pytest.mark.asyncio
    async def test_conversion_preserves_dc_id(self, account_id, tmp_path):
        """DC ID should be preserved through TDesktop -> TelegramClient."""
        tdata_path = str(TDATAS_DIR / account_id / "tdata")
        tdesk = TDesktop(tdata_path)

        original_dc = int(tdesk.mainAccount.MainDcId)

        session_path = str(tmp_path / f"test_dc_{account_id}")
        client = await tdesk.ToTelethon(
            session=session_path,
            flag=UseCurrentSession,
            api=API.TelegramDesktop,
        )

        assert client.session.dc_id == original_dc

    @pytest.mark.asyncio
    async def test_roundtrip_tdesk_telethon_tdesk(self, account_id, tmp_path):
        """TDesktop -> TelegramClient -> TDesktop should preserve auth key."""
        tdata_path = str(TDATAS_DIR / account_id / "tdata")
        tdesk = TDesktop(tdata_path)

        original_key = tdesk.mainAccount.authKey.key
        original_dc = int(tdesk.mainAccount.MainDcId)
        original_uid = tdesk.mainAccount.UserId

        session_path = str(tmp_path / f"test_rt_{account_id}")
        client = await tdesk.ToTelethon(
            session=session_path,
            flag=UseCurrentSession,
            api=API.TelegramDesktop,
        )

        tdesk2 = await TDesktop.FromTelethon(
            client,
            flag=UseCurrentSession,
            api=API.TelegramDesktop,
        )

        assert tdesk2.isLoaded()
        assert tdesk2.mainAccount is not None
        assert tdesk2.mainAccount.authKey.key == original_key
        assert int(tdesk2.mainAccount.MainDcId) == original_dc
        assert tdesk2.mainAccount.UserId == original_uid

    @pytest.mark.asyncio
    async def test_conversion_with_json_api(self, account_id, tmp_path):
        """Convert using the API params from the JSON file."""
        tdata_path = str(TDATAS_DIR / account_id / "tdata")
        jdata = _load_json(account_id)

        api_map = {
            2040: API.TelegramDesktop,
            6: API.TelegramAndroid,
            21724: API.TelegramAndroidX,
            10840: API.TelegramIOS,
            2834: API.TelegramMacOS,
            2496: API.TelegramWeb_Z,
        }
        api_cls = api_map[jdata["app_id"]]

        tdesk = TDesktop(tdata_path)
        session_path = str(tmp_path / f"test_japi_{account_id}")
        client = await tdesk.ToTelethon(
            session=session_path,
            flag=UseCurrentSession,
            api=api_cls,
        )
        assert client is not None
        assert client.session.auth_key is not None


# ===================================================================
# 9. TELEGRAM CLIENT CONSTRUCTOR TESTS
# ===================================================================


class TestTelegramClientConstructor:
    def test_default_api(self):
        client = TelegramClient()
        assert client.api_id == API.TelegramDesktop.api_id
        assert client.api_hash == API.TelegramDesktop.api_hash

    def test_custom_api_id(self):
        client = TelegramClient(None, 1234, "testhash")
        assert client.api_id == 1234

    def test_api_object(self):
        client = TelegramClient(None, api=API.TelegramAndroid)
        assert client.api_id == API.TelegramAndroid.api_id

    def test_api_instance(self):
        api = API.TelegramIOS()
        client = TelegramClient(None, api=api)
        assert client.api_id == API.TelegramIOS.api_id


# ===================================================================
# 10. TDATA SAVE/LOAD ROUNDTRIP
# ===================================================================


class TestTDataSaveLoad:
    """Test saving tdata and reloading it."""

    @pytest.mark.asyncio
    async def test_save_and_reload(self, tmp_path):
        """Load tdata, convert, save, and reload."""
        account_id = ACCOUNT_IDS[0]
        tdata_path = str(TDATAS_DIR / account_id / "tdata")
        tdesk = TDesktop(tdata_path)

        original_key = tdesk.mainAccount.authKey.key
        original_uid = tdesk.mainAccount.UserId

        # Convert to TelegramClient and back
        session_path = str(tmp_path / "save_test")
        client = await tdesk.ToTelethon(
            session=session_path,
            flag=UseCurrentSession,
            api=API.TelegramDesktop,
        )

        tdesk2 = await TDesktop.FromTelethon(
            client, flag=UseCurrentSession, api=API.TelegramDesktop
        )

        # Save to new tdata path
        save_path = str(tmp_path / "saved_tdata")
        os.makedirs(save_path, exist_ok=True)
        tdesk2.SaveTData(save_path)

        # Reload
        tdesk3 = TDesktop(save_path)
        assert tdesk3.isLoaded()
        assert tdesk3.mainAccount.authKey.key == original_key
        assert tdesk3.mainAccount.UserId == original_uid


# ===================================================================
# 11. EDGE CASES AND ERROR HANDLING
# ===================================================================


class TestEdgeCases:
    def test_tdesktop_empty_init(self):
        tdesk = TDesktop()
        assert not tdesk.isLoaded()
        assert tdesk.accountsCount == 0

    def test_tdesktop_invalid_path(self):
        with pytest.raises(BaseException):
            TDesktop("/nonexistent/path/tdata")

    def test_api_data_requires_id_and_hash(self):
        with pytest.raises(BaseException):
            APIData()  # no api_id or api_hash

    def test_login_flag_hierarchy(self):
        assert issubclass(UseCurrentSession, LoginFlag)
        assert issubclass(CreateNewSession, LoginFlag)
        assert issubclass(LoginFlag, int)

    def test_device_info_str(self):
        d = DeviceInfo("Model X", "v1.0")
        assert str(d) == "Model X v1.0"


# ===================================================================
# 12. LIVE ACCOUNT TESTS (network required)
# ===================================================================

# Mark all live tests so they can be run selectively:
#   pytest -m live
# or skipped:
#   pytest -m "not live"


def _session_path(account_id: str) -> str:
    return str(SESSIONS_DIR / account_id)


def _tdata_path(account_id: str) -> str:
    return str(TDATAS_DIR / account_id / "tdata")


def _safe_print(text: str) -> None:
    """Print text safely on consoles that don't support full Unicode (e.g. cp1251)."""
    try:
        print(text)
    except UnicodeEncodeError:
        print(text.encode("ascii", errors="replace").decode("ascii"))


class TestLiveSessionAccounts:
    """Connect using .session files, verify authorization, print account info."""

    @pytest.fixture(params=ACCOUNT_IDS)
    def account_id(self, request):
        return request.param

    @pytest.mark.live
    @pytest.mark.asyncio
    async def test_session_connect_and_get_me(self, account_id):
        """Load session, connect, call get_me, print account info."""
        from telethon.errors import UserDeactivatedBanError

        client = await TelegramClient.FromSessionJson(_session_path(account_id))
        try:
            await client.connect()

            try:
                authorized = await client.is_user_authorized()
            except UserDeactivatedBanError:
                pytest.skip(f"Session {account_id} is deactivated/banned")
                return

            if not authorized:
                pytest.skip(f"Session {account_id} is not authorized")
                return

            me = await client.get_me()
            assert me is not None, f"get_me() returned None for {account_id}"

            _safe_print(f"\n{'=' * 60}")
            _safe_print(f"  Account ID : {me.id}")
            _safe_print(f"  Phone      : +{me.phone}")
            _safe_print(f"  First Name : {me.first_name}")
            _safe_print(f"  Last Name  : {me.last_name or ''}")
            _safe_print(f"  Username   : @{me.username or 'N/A'}")
            _safe_print(f"  Premium    : {me.premium}")
            _safe_print(f"  DC ID      : {client.session.dc_id}")
            _safe_print(f"  Source     : session file ({account_id})")
            _safe_print(f"{'=' * 60}")
        finally:
            await client.disconnect()

    @pytest.mark.live
    @pytest.mark.asyncio
    async def test_session_get_sessions(self, account_id):
        """Load session, connect, get all active sessions and print them."""
        from telethon.errors import UserDeactivatedBanError

        client = await TelegramClient.FromSessionJson(_session_path(account_id))
        try:
            await client.connect()

            try:
                authorized = await client.is_user_authorized()
            except UserDeactivatedBanError:
                pytest.skip(f"Session {account_id} is deactivated/banned")
                return

            if not authorized:
                pytest.skip(f"Session {account_id} not authorized")
                return

            sessions = await client.GetSessions()
            assert sessions is not None
            assert len(sessions.authorizations) > 0

            _safe_print(f"\n--- Sessions for account {account_id} (from .session) ---")
            for auth in sessions.authorizations:
                tag = "(current)" if auth.current else "         "
                _safe_print(
                    f"  {tag} {auth.device_model} | "
                    f"{auth.platform} | {auth.system_version} | "
                    f"api_id={auth.api_id} | {auth.app_name}"
                )
        finally:
            await client.disconnect()


class TestLiveTDataAccounts:
    """Connect using tdata folders, verify authorization, print account info."""

    @pytest.fixture(params=ACCOUNT_IDS)
    def account_id(self, request):
        return request.param

    @pytest.mark.live
    @pytest.mark.asyncio
    async def test_tdata_connect_and_get_me(self, account_id, tmp_path):
        """Load tdata, convert to TelegramClient, connect, print account info."""
        from telethon.errors import UserDeactivatedBanError

        tdesk = TDesktop(_tdata_path(account_id))
        assert tdesk.isLoaded(), f"Failed to load tdata for {account_id}"

        session_path = str(tmp_path / f"live_tdata_{account_id}")
        client = await tdesk.ToTelethon(
            session=session_path,
            flag=UseCurrentSession,
            api=API.TelegramDesktop,
        )
        try:
            await client.connect()

            try:
                authorized = await client.is_user_authorized()
            except UserDeactivatedBanError:
                pytest.skip(f"tdata {account_id} is deactivated/banned")
                return

            if not authorized:
                pytest.skip(f"tdata {account_id} session is not authorized")
                return

            me = await client.get_me()
            assert me is not None, f"get_me() returned None for tdata {account_id}"

            _safe_print(f"\n{'=' * 60}")
            _safe_print(f"  Account ID : {me.id}")
            _safe_print(f"  Phone      : +{me.phone}")
            _safe_print(f"  First Name : {me.first_name}")
            _safe_print(f"  Last Name  : {me.last_name or ''}")
            _safe_print(f"  Username   : @{me.username or 'N/A'}")
            _safe_print(f"  Premium    : {me.premium}")
            _safe_print(f"  DC ID      : {client.session.dc_id}")
            _safe_print(f"  Source     : tdata folder ({account_id})")
            _safe_print(f"{'=' * 60}")
        finally:
            await client.disconnect()

    @pytest.mark.live
    @pytest.mark.asyncio
    async def test_tdata_get_sessions(self, account_id, tmp_path):
        """Load tdata, convert, connect, get all active sessions."""
        from telethon.errors import UserDeactivatedBanError

        tdesk = TDesktop(_tdata_path(account_id))
        assert tdesk.isLoaded()

        session_path = str(tmp_path / f"live_tdata_gs_{account_id}")
        client = await tdesk.ToTelethon(
            session=session_path,
            flag=UseCurrentSession,
            api=API.TelegramDesktop,
        )
        try:
            await client.connect()

            try:
                authorized = await client.is_user_authorized()
            except UserDeactivatedBanError:
                pytest.skip(f"tdata {account_id} is deactivated/banned")
                return

            if not authorized:
                pytest.skip(f"tdata {account_id} not authorized")
                return

            sessions = await client.GetSessions()
            assert sessions is not None
            assert len(sessions.authorizations) > 0

            _safe_print(f"\n--- Sessions for account {account_id} (from tdata) ---")
            for auth in sessions.authorizations:
                tag = "(current)" if auth.current else "         "
                _safe_print(
                    f"  {tag} {auth.device_model} | "
                    f"{auth.platform} | {auth.system_version} | "
                    f"api_id={auth.api_id} | {auth.app_name}"
                )
        finally:
            await client.disconnect()


class TestLiveCreateNewSession:
    """Try creating a new session from an existing authorized session."""

    @pytest.mark.live
    @pytest.mark.asyncio
    async def test_create_new_session_from_tdata(self, tmp_path):
        """Load first tdata account, try CreateNewSession (QR login)."""
        from telethon.errors import UserDeactivatedBanError

        account_id = ACCOUNT_IDS[0]
        tdesk = TDesktop(_tdata_path(account_id))
        assert tdesk.isLoaded()

        # First get a working client with UseCurrentSession
        session_path = str(tmp_path / f"source_{account_id}")
        source_client = await tdesk.ToTelethon(
            session=session_path,
            flag=UseCurrentSession,
            api=API.TelegramDesktop,
        )
        try:
            await source_client.connect()

            try:
                authorized = await source_client.is_user_authorized()
            except UserDeactivatedBanError:
                pytest.skip("Source session is deactivated/banned")
                return

            if not authorized:
                pytest.skip(
                    "Source session not authorized, cannot test CreateNewSession"
                )
                return

            me = await source_client.get_me()
            _safe_print("\n--- CreateNewSession test ---")
            _safe_print(f"  Source account: {me.id} (+{me.phone})")

            # Attempt QR login to create a new session
            new_session_path = str(tmp_path / f"new_{account_id}")
            try:
                new_client = await source_client.QRLoginToNewClient(
                    session=new_session_path,
                    api=API.TelegramDesktop,
                )
                new_me = await new_client.get_me()
                _safe_print("  New session created successfully!")
                _safe_print(f"  New session account: {new_me.id}")
                _safe_print(f"  New session DC: {new_client.session.dc_id}")
                assert new_me.id == me.id, "New session should be for the same account"
                await new_client.disconnect()
            except Exception as e:
                _safe_print(
                    f"  CreateNewSession failed (expected in CI): {type(e).__name__}: {e}"
                )
                # QR login requires interactive approval, so failure is expected
                # in automated tests - we just verify the attempt was made
        finally:
            await source_client.disconnect()

    @pytest.mark.live
    @pytest.mark.asyncio
    async def test_save_session_json_with_live_info(self, tmp_path):
        """Load session, connect, save with fetch_user_info=True, verify JSON."""
        from telethon.errors import UserDeactivatedBanError

        account_id = ACCOUNT_IDS[0]
        client = await TelegramClient.FromSessionJson(_session_path(account_id))
        try:
            await client.connect()

            try:
                authorized = await client.is_user_authorized()
            except UserDeactivatedBanError:
                pytest.skip("Session is deactivated/banned")
                return

            if not authorized:
                pytest.skip("Session not authorized")
                return

            out_base = str(tmp_path / f"live_export_{account_id}")
            s_path, j_path = await client.SaveSessionJson(
                out_base, fetch_user_info=True
            )

            assert os.path.isfile(s_path)
            assert os.path.isfile(j_path)

            with open(j_path, "r", encoding="utf-8") as f:
                data = json.load(f)

            _safe_print("\n--- SaveSessionJson with fetch_user_info ---")
            _safe_print(f"  session_file : {data.get('session_file')}")
            _safe_print(f"  id           : {data.get('id')}")
            _safe_print(f"  phone        : {data.get('phone')}")
            _safe_print(f"  username     : {data.get('username')}")
            _safe_print(f"  is_premium   : {data.get('is_premium')}")
            _safe_print(f"  app_id       : {data.get('app_id')}")
            _safe_print(f"  device       : {data.get('device')}")

            # The fetched info should be populated
            assert data.get("id") is not None, "id should be set with fetch_user_info"
            assert data.get("phone") is not None, (
                "phone should be set with fetch_user_info"
            )
        finally:
            await client.disconnect()



"""Post-login consistency checks for opentele2.

After connecting and authenticating, official Telegram clients typically
call a series of ``help.*`` methods that the server expects.  Skipping
them can look suspicious.  This module provides helpers that mimic the
official post-login behaviour.

Usage::

    from opentele2.consistency import ConsistencyChecker

    checker = ConsistencyChecker(client)
    report = await checker.run_all()
    # report is a ConsistencyReport with per-check results
"""

from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from typing import Any, List, TYPE_CHECKING

if TYPE_CHECKING:
    pass  # avoid circular imports

__all__ = [
    "ConsistencyChecker",
    "ConsistencyReport",
    "CheckResult",
]


@dataclass
class CheckResult:
    """Result of a single consistency check."""

    name: str
    passed: bool
    detail: str = ""
    server_response: Any = None


@dataclass
class ConsistencyReport:
    """Aggregated results from all consistency checks."""

    checks: List[CheckResult] = field(default_factory=list)

    @property
    def all_passed(self) -> bool:
        return all(c.passed for c in self.checks)

    @property
    def summary(self) -> str:
        total = len(self.checks)
        passed = sum(1 for c in self.checks if c.passed)
        lines = [f"Consistency: {passed}/{total} checks passed"]
        for c in self.checks:
            status = "OK" if c.passed else "FAIL"
            lines.append(f"  [{status}] {c.name}: {c.detail}")
        return "\n".join(lines)

    def __str__(self) -> str:
        return self.summary


class ConsistencyChecker:
    """Runs post-login checks that mimic official client behaviour.

    All methods are safe to call — they are read-only requests that
    official clients routinely make.

    Args:
        client: A connected and authorized ``TelegramClient`` instance.
        auto_warn: If True (default), emit warnings for failed checks.
    """

    def __init__(self, client, *, auto_warn: bool = True):
        self._client = client
        self._auto_warn = auto_warn

    async def run_all(self) -> ConsistencyReport:
        """Run all consistency checks and return a report."""
        report = ConsistencyReport()

        checks = [
            self.check_get_config,
            self.check_current_session,
            self.check_layer_match,
            self.check_lang_pack,
            self.check_app_update,
            self.check_terms_of_service,
        ]

        for check_fn in checks:
            try:
                result = await check_fn()
                report.checks.append(result)
            except Exception as e:
                report.checks.append(
                    CheckResult(
                        name=check_fn.__name__.replace("check_", ""),
                        passed=False,
                        detail=f"Exception: {e}",
                    )
                )

        if self._auto_warn and not report.all_passed:
            warnings.warn(
                f"opentele2 consistency check issues:\n{report.summary}",
                stacklevel=2,
            )

        return report

    async def check_get_config(self) -> CheckResult:
        """Call ``help.getConfig`` — every official client does this on connect.

        Verifies the server returns a valid config without errors.
        This is the single most important post-connection call.
        """
        from telethon import functions

        try:
            config = await self._client(functions.help.GetConfigRequest())
            dc_id = getattr(config, "this_dc", None)
            return CheckResult(
                name="get_config",
                passed=True,
                detail=f"Config OK (DC {dc_id})",
                server_response=config,
            )
        except Exception as e:
            return CheckResult(
                name="get_config",
                passed=False,
                detail=f"Server rejected getConfig: {e}",
            )

    async def check_current_session(self) -> CheckResult:
        """Verify the current session via ``account.getAuthorizations``.

        Checks that the session is seen as an official app.
        """
        try:
            auth = await self._client.GetCurrentSession()
            if auth is None:
                return CheckResult(
                    name="current_session",
                    passed=False,
                    detail="Could not retrieve current session",
                )

            is_official = bool(auth.official_app)
            detail = (
                f"api_id={auth.api_id}, official={is_official}, "
                f"device='{auth.device_model}', "
                f"app='{auth.app_name} {auth.app_version}'"
            )
            return CheckResult(
                name="current_session",
                passed=is_official,
                detail=detail,
                server_response=auth,
            )
        except Exception as e:
            return CheckResult(
                name="current_session",
                passed=False,
                detail=f"Error: {e}",
            )

    async def check_layer_match(self) -> CheckResult:
        """Check if our layer matches what the server expects.

        We cannot directly query the server's expected layer, but we can
        check that ``help.getConfig`` did not return an error and that
        the session's app matches.
        """
        from .fingerprint import LAYER, _detect_telethon_layer

        telethon_layer = _detect_telethon_layer()
        diff = abs(telethon_layer - LAYER)

        passed = diff <= 15
        detail = (
            f"Telethon layer={telethon_layer}, "
            f"official schema layer={LAYER}, "
            f"diff={diff}"
        )
        return CheckResult(name="layer_match", passed=passed, detail=detail)

    async def check_lang_pack(self) -> CheckResult:
        """Verify that the lang_pack used at initConnection is valid.

        Calls ``langpack.getLanguages`` which only succeeds for valid
        lang_pack values.
        """
        from telethon import functions

        # Retrieve the lang_pack from the client's API settings
        lang_pack = getattr(self._client, "_sender", None)
        if lang_pack is not None:
            lang_pack = getattr(lang_pack, "_init_request", None)
            if lang_pack is not None:
                lang_pack = getattr(lang_pack, "lang_pack", "")
            else:
                lang_pack = ""
        else:
            lang_pack = ""

        if not lang_pack:
            return CheckResult(
                name="lang_pack",
                passed=True,
                detail="lang_pack is empty (web client mode)",
            )

        try:
            languages = await self._client(
                functions.langpack.GetLanguagesRequest(lang_pack=lang_pack)
            )
            return CheckResult(
                name="lang_pack",
                passed=True,
                detail=f"lang_pack '{lang_pack}' valid ({len(languages)} languages)",
                server_response=languages,
            )
        except Exception as e:
            return CheckResult(
                name="lang_pack",
                passed=False,
                detail=f"lang_pack '{lang_pack}' rejected: {e}",
            )

    async def check_app_update(self) -> CheckResult:
        """Call ``help.getAppUpdate`` — official clients do this periodically.

        A successful call (or ``noAppUpdate``) means our app_version is
        accepted by the server.
        """
        from telethon import functions, types

        try:
            # help.getAppUpdate requires a source parameter
            result = await self._client(functions.help.GetAppUpdateRequest(source=""))
            if isinstance(result, types.help.NoAppUpdate):
                return CheckResult(
                    name="app_update",
                    passed=True,
                    detail="No update available (app_version accepted)",
                    server_response=result,
                )
            else:
                version = getattr(result, "version", "?")
                return CheckResult(
                    name="app_update",
                    passed=True,
                    detail=f"Update available: v{version} (but request succeeded)",
                    server_response=result,
                )
        except Exception as e:
            return CheckResult(
                name="app_update",
                passed=False,
                detail=f"Error calling getAppUpdate: {e}",
            )

    async def check_terms_of_service(self) -> CheckResult:
        """Call ``help.getTermsOfServiceUpdate`` — official clients check this.

        Verifies the server doesn't flag us.
        """
        from telethon import functions

        try:
            result = await self._client(functions.help.GetTermsOfServiceUpdateRequest())
            return CheckResult(
                name="terms_of_service",
                passed=True,
                detail="ToS check OK",
                server_response=result,
            )
        except Exception as e:
            return CheckResult(
                name="terms_of_service",
                passed=False,
                detail=f"Error: {e}",
            )



from __future__ import annotations

import os
import platform
from typing import Any, List, Dict, Tuple, Type, TypeVar, Union, Optional

from .devices import (
    AndroidDevice,
    IOSDevice,
    macOSDevice,
    SystemInfo,
    WindowsDevice,
    LinuxDevice,
    WebBrowserDevice,
)
from .exception import Expects, NoInstanceMatched
from .utils import BaseMetaClass, BaseObject, sharemethod

__all__ = [
    "APIData",
    "API",
    "LoginFlag",
    "UseCurrentSession",
    "CreateNewSession",
]

_T = TypeVar("_T")

_X64_ARCHES = ("x86_64", "AMD64")
_X86_ARCHES = ("i386", "i686", "x86")

_DEFAULT_EXTRA_FIELDS = {
    "twoFA": None,
    "role": "",
    "id": None,
    "phone": None,
    "username": None,
    "date_of_birth": None,
    "date_of_birth_integrity": None,
    "is_premium": False,
    "has_profile_pic": False,
    "spamblock": None,
    "register_time": None,
    "last_check_time": None,
    "avatar": None,
    "first_name": "",
    "last_name": "",
    "sex": None,
    "proxy": None,
    "ipv6": False,
    "session_file": "",
}


def _coalesce(val, default):
    return val if val is not None else default


class BaseAPIMetaClass(BaseMetaClass):
    def __new__(
        cls: Type[_T], clsName: str, bases: Tuple[type], attrs: Dict[str, Any]
    ) -> _T:
        result = super().__new__(cls, clsName, bases, attrs)
        result._clsMakePID()  # type: ignore
        result.__str__ = BaseAPIMetaClass.__str__  # type: ignore

        return result

    @sharemethod
    def __str__(glob) -> str:
        if isinstance(glob, type):
            cls = glob
            result = f"{cls.__name__} {{\n"
        else:
            cls = glob.__class__
            result = f"{cls.__name__}() = {{\n"

        for attr, val in glob.__dict__.items():
            if (
                attr.startswith(f"_{cls.__base__.__name__}__")
                or attr.startswith(f"_{cls.__name__}__")
                or attr.startswith("__")
                and attr.endswith("__")
                or type(val) is classmethod
                or callable(val)
            ):
                continue

            result += f"    {attr}: {val}\n"

        return result + "}"


class APIData(object, metaclass=BaseAPIMetaClass):
    """API configuration to connect to TelegramClient and TDesktop."""

    CustomInitConnectionList: List[Union[Type[APIData], APIData]] = []

    api_id: int = None  # type: ignore
    api_hash: str = None  # type: ignore
    device_model: str = None  # type: ignore
    system_version: str = None  # type: ignore
    app_version: str = None  # type: ignore
    lang_code: str = None  # type: ignore
    system_lang_code: str = None  # type: ignore
    lang_pack: str = None  # type: ignore

    def __init__(
        self,
        api_id: int = None,
        api_hash: str = None,
        device_model: str = None,
        system_version: str = None,
        app_version: str = None,
        lang_code: str = None,
        system_lang_code: str = None,
        lang_pack: str = None,
    ) -> None:
        Expects(
            (self.__class__ != APIData)
            or (api_id is not None and api_hash is not None),
            NoInstanceMatched("No instace of API matches the arguments"),
        )

        cls = self.get_cls()

        self.api_id = _coalesce(api_id, cls.api_id)
        self.api_hash = _coalesce(api_hash, cls.api_hash)
        self.device_model = _coalesce(device_model, cls.device_model)
        self.system_version = _coalesce(system_version, cls.system_version)
        self.app_version = _coalesce(app_version, cls.app_version)
        self.system_lang_code = _coalesce(system_lang_code, cls.system_lang_code)
        self.lang_pack = _coalesce(lang_pack, cls.lang_pack)
        self.lang_code = _coalesce(lang_code, cls.lang_code)

        if self.device_model is None:
            system = platform.uname()
            if system.machine in _X64_ARCHES:
                self.device_model = "PC 64bit"
            elif system.machine in _X86_ARCHES:
                self.device_model = "PC 32bit"
            else:
                self.device_model = system.machine

        self._makePID()

    @sharemethod
    def copy(glob: Union[Type[_T], _T] = _T) -> _T:  # type: ignore
        cls = glob if isinstance(glob, type) else glob.__class__

        return cls(
            glob.api_id,  # type: ignore
            glob.api_hash,  # type: ignore
            glob.device_model,  # type: ignore
            glob.system_version,  # type: ignore
            glob.app_version,  # type: ignore
            glob.lang_code,  # type: ignore
            glob.system_lang_code,  # type: ignore
            glob.lang_pack,  # type: ignore
        )  # type: ignore

    @sharemethod
    def get_cls(glob: Union[Type[_T], _T]) -> Type[_T]:  # type: ignore
        return glob if isinstance(glob, type) else glob.__class__

    @sharemethod
    def destroy(glob: Union[Type[_T], _T]):  # type: ignore
        if isinstance(glob, type):
            return

    @classmethod
    def _web_generate(cls: Type[_T], unique_id: str = None, variant: str = "z") -> _T:
        deviceInfo = WebBrowserDevice.RandomDevice(unique_id, variant=variant)
        return cls(device_model=deviceInfo.model, system_version=deviceInfo.version)

    @staticmethod
    def from_json(data: dict) -> APIData:
        """Create an APIData instance from a session JSON dictionary.

        Args:
            data: Dictionary loaded from a .json session metadata file.
                  Expected fields: app_id, app_hash, device, sdk, app_version,
                  lang_pack, lang_code, and either system_lang_pack or
                  system_lang_code.

        Returns:
            APIData instance with all fields populated.
        """
        api_id = data.get("app_id")
        api_hash = data.get("app_hash")
        system_lang_code = data.get("system_lang_code") or data.get(
            "system_lang_pack", "en"
        )

        return APIData(
            api_id=api_id,
            api_hash=api_hash,
            device_model=data.get("device"),
            system_version=data.get("sdk"),
            app_version=data.get("app_version"),
            lang_code=data.get("lang_code", "en"),
            system_lang_code=system_lang_code,
            lang_pack=data.get("lang_pack", ""),
        )

    def to_json(self, extra: dict = None) -> dict:
        """Serialize this APIData to a session JSON dictionary.

        Args:
            extra: Optional dictionary of extra fields to include
                   (twoFA, phone, username, etc.). If None, defaults
                   are used for all extra fields.

        Returns:
            Dictionary ready for json.dump().
        """
        result = {
            "app_id": self.api_id,
            "app_hash": self.api_hash,
            "device": self.device_model,
            "sdk": self.system_version,
            "app_version": self.app_version,
            "system_lang_pack": self.system_lang_code,
            "system_lang_code": self.system_lang_code,
            "lang_pack": self.lang_pack,
            "lang_code": self.lang_code,
        }
        defaults = dict(_DEFAULT_EXTRA_FIELDS)
        if extra:
            defaults.update(extra)
        result.update(defaults)
        return result

    def __eq__(self, __o: APIData) -> bool:
        if not isinstance(__o, APIData):
            return False
        return self.pid == __o.pid

    def __del__(self):
        self.destroy()

    @classmethod
    def _makePIDEnsure(cls) -> int:
        while True:
            pid = int.from_bytes(os.urandom(8), "little")
            if cls.findData(pid) is None:
                break
        return pid

    @classmethod
    def _clsMakePID(cls: Type[APIData]):
        cls.pid = cls._makePIDEnsure()
        cls.CustomInitConnectionList.append(cls)

    def _makePID(self):
        self.pid = self.get_cls()._makePIDEnsure()
        self.get_cls().CustomInitConnectionList.append(self)

    @classmethod
    def Generate(cls: Type[_T], unique_id: str = None) -> _T:
        """Generate random device model and system version.

        Args:
            unique_id: The unique ID to generate - ensures same data each time.
                       If not set, data is randomized on each run.

        Raises:
            NotImplementedError: Not supported for web browser APIs.
        """
        if cls == API.TelegramAndroid or cls == API.TelegramAndroidX:
            deviceInfo = AndroidDevice.RandomDevice(unique_id)

        elif cls == API.TelegramIOS:
            deviceInfo = IOSDevice.RandomDevice(unique_id)

        elif cls == API.TelegramMacOS:
            deviceInfo = macOSDevice.RandomDevice(unique_id)

        elif cls in (API.TelegramWeb_Z, API.TelegramWeb_A):
            deviceInfo = WebBrowserDevice.RandomDevice(unique_id, variant="z")

        elif cls == API.TelegramWeb_K:
            deviceInfo = WebBrowserDevice.RandomDevice(unique_id, variant="k")

        elif cls == API.Webogram:
            deviceInfo = WebBrowserDevice.RandomDevice(unique_id, variant="k")

        else:
            raise NotImplementedError(
                f"{cls.__name__} device not supported for randomize yet"
            )

        return cls(device_model=deviceInfo.model, system_version=deviceInfo.version)

    @classmethod
    def findData(cls: Type[_T], pid: int) -> Optional[_T]:
        for x in cls.CustomInitConnectionList:  # type: ignore
            if x.pid == pid:
                return x
        return None


class API(BaseObject):
    """Built-in templates for official Telegram APIs.

    Uses official API credentials, making sessions indistinguishable from
    official apps. Supports the lang_pack parameter for official apps only.
    """

    class TelegramDesktop(APIData):
        """Official Telegram for Desktop (Windows, macOS, Linux)."""

        api_id = 2040
        api_hash = "b18441a1ff607e10a989891a5462e627"
        device_model = "Desktop"
        system_version = "Windows 11"
        app_version = "5.12.3 x64"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = "tdesktop"

        @classmethod
        def Generate(cls: Type[_T], system: str = None, unique_id: str = None) -> _T:
            """Generate random TelegramDesktop device data.

            Args:
                system: Target OS - "windows", "macos", or "linux". Random if None.
                unique_id: Ensures deterministic generation per ID.
            """
            validList = ["windows", "macos", "linux"]
            if system is None or system not in validList:
                system = SystemInfo._hashtovalue(
                    SystemInfo._strtohashid(unique_id), validList
                )

            system = system.lower()

            if system == "windows":
                deviceInfo = WindowsDevice.RandomDevice(unique_id)
            elif system == "macos":
                deviceInfo = macOSDevice.RandomDevice(unique_id)
            else:
                deviceInfo = LinuxDevice.RandomDevice(unique_id)

            return cls(device_model=deviceInfo.model, system_version=deviceInfo.version)

    class TelegramAndroid(APIData):
        """Official Telegram for Android."""

        api_id = 6
        api_hash = "eb06d4abfb49dc3eeb1aeb98ae0f581e"
        device_model = "Samsung Galaxy S24 Ultra"
        system_version = "SDK 35"
        app_version = "12.3.0"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = "android"

    class TelegramAndroidX(APIData):
        """Official TelegramX for Android."""

        api_id = 21724
        api_hash = "3e0cb5efcd52300aec5994fdfc5bdc16"
        device_model = "Samsung Galaxy S24 Ultra"
        system_version = "SDK 35"
        app_version = "12.3.0"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = "android"

    class TelegramIOS(APIData):
        """Official Telegram for iOS."""

        api_id = 10840
        api_hash = "33c45224029d59cb3ad0c16134215aeb"
        device_model = "iPhone 16 Pro Max"
        system_version = "26.2"
        app_version = "12.3"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = "ios"

    class TelegramMacOS(APIData):
        """Official Telegram-Swift for macOS."""

        api_id = 2834
        api_hash = "68875f756c9b437a8b916ca3de215815"
        device_model = "MacBook Pro"
        system_version = "macOS 26.2"
        app_version = "12.3"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = "macos"

    class TelegramWeb_Z(APIData):
        """Official Telegram WebZ for Browsers."""

        api_id = 2496
        api_hash = "8da85b0d5bfe62527e5b244c209159c3"
        device_model = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/144.0.0.0 Safari/537.36"
        system_version = "Windows"
        app_version = "5.0.0 Z"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = ""

        @classmethod
        def Generate(cls: Type[_T], unique_id: str = None) -> _T:
            return cls._web_generate(unique_id, variant="z")

    class TelegramWeb_A(APIData):
        """Official Telegram WebA for Browsers."""

        api_id = 2496
        api_hash = "8da85b0d5bfe62527e5b244c209159c3"
        device_model = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/144.0.0.0 Safari/537.36"
        system_version = "Windows"
        app_version = "5.0.0 A"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = ""

        @classmethod
        def Generate(cls: Type[_T], unique_id: str = None) -> _T:
            return cls._web_generate(unique_id, variant="z")

    class TelegramWeb_K(APIData):
        """Official Telegram WebK (legacy) for Browsers."""

        api_id = 2496
        api_hash = "8da85b0d5bfe62527e5b244c209159c3"
        device_model = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/144.0.0.0 Safari/537.36"
        system_version = "Win32"
        app_version = "1.4.2 K"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = "macos"

        @classmethod
        def Generate(cls: Type[_T], unique_id: str = None) -> _T:
            return cls._web_generate(unique_id, variant="k")

    class Webogram(APIData):
        """Old Telegram for Browsers (legacy)."""

        api_id = 2496
        api_hash = "8da85b0d5bfe62527e5b244c209159c3"
        device_model = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/144.0.0.0 Safari/537.36"
        system_version = "Win32"
        app_version = "0.7.0"
        lang_code = "en"
        system_lang_code = "en-US"
        lang_pack = ""

        @classmethod
        def Generate(cls: Type[_T], unique_id: str = None) -> _T:
            return cls._web_generate(unique_id, variant="k")


class LoginFlag(int):
    """Flag for converting sessions between TDesktop and TelegramClient."""


class UseCurrentSession(LoginFlag):
    """Use the current session - reads the 256-byte AuthKey and converts it directly.

    Warning: Only use the same consistent API throughout the session.
    """


class CreateNewSession(LoginFlag):
    """Create a new session via QR code login using the current session for authorization.

    Safe to use with any API, even different from the original session's API.
    """



from __future__ import annotations

from ..utils import BaseObject
from ..exception import (
    Expects,
    LoginFlagInvalid,
    OpenTeleException,
    QDataStreamFailed,
    TDataBadDecryptKey,
    TDataSaveFailed,
    TDesktopHasNoAccount,
    TDesktopNotLoaded,
)
from ..api import API, APIData, CreateNewSession, LoginFlag, UseCurrentSession
from . import shared as td
from .. import tl
from typing import List, Optional, Type, Union
from ctypes import sizeof, c_int32 as int32
from ..qt_compat import QByteArray, QDataStream
import asyncio
import typing

import logging
from telethon.network.connection.connection import Connection
from telethon.network.connection.tcpfull import ConnectionTcpFull
from telethon.sessions.abstract import Session

_PERF_LOCAL_KEY = bytes.fromhex(
    "d8745944519e0d2d71309d6c8d272dc64948f5e3eba7685324d5c691ad810c20"
    "3b31d19d29aed6ac33c014be6e09843293f6fa32dbe42b6a04e00481fae99511"
    "4caf6342bd98e96d293dd062c458689b3abd23a5cf230c75527c05bf5f90f38c"
    "d93952cf61aaac1cfeaae4608592e363ded35f8d8c45234def53231decb35592"
    "afc40d0601bbed11090969f74d9ab0cc97827546f441242d2cfb8e05a0610e97"
    "669c0da1adccb56e39e10c69e2942387ff4922f8c55dcb8890e345ef318266f4"
    "b3831430ea210c863c17624c0494cdead81f523430b5f74c15da323d766bd01c"
    "b5b88b9d2a731f6d853380ad306a8647fa614cc4017f08902c1e1f997ee12e3c"
)

_PERF_PASSCODE_KEY_SALT = QByteArray(
    bytes.fromhex("aed1e082994281d975760e729560d2c8d008f2a9dd3ff4d83245e22eedb66716")
)

_PERF_PASSCODE_KEY = bytes.fromhex(
    "273e6483b313c9dbc4acd9173864c0422f172881bfe1c6649ca5538654d0bd6e"
    "bcfbcc8dabbb669717ce53b41ca8afbf9d15c23fa2b06a6e166cc61f62e99805"
    "58cd42cf1022e05b466811ea29e9359fc1f94668e5c85155bce638fe7aa42aa8"
    "80ca83e0d2e534198a0a890139c8f679a57d3ba06d75fe5faa6e250101a58bc6"
    "e42b9696604de1e34ae40f6bba144a284f3ad8843253ec9b3971863a2c409208"
    "c2563967b3587e509b42a42a6040d23ff696ad552a240084fa3f950240f499b2"
    "3cd6d27e7010cbde07daae0667a7df8a5115a60b265c58f5d9291f7a98fc3c60"
    "1e2a4a32f1881b8218c855239d7b532959609e6ab52e48ad691c2583b566c8f9"
)

_PERF_PASSCODE_KEY_ENCRYPTED = QByteArray(
    bytes.fromhex(
        "97df0cd2e3109149b77b5287994d9c1ca240c51e87488e79dd029bea65fb9d27"
        "89bb5abcfe65e871d752bd938d83313c794c8993a734ce1216f2e660473f3143"
        "af9a333610a17995876e1721ce1f611d1c69d8c1a2f59f9493119704274e2cb5"
        "f36c20df439d156deff7a34371dc44bc86f8730cebf9b028eb7a1ed6621d99ad"
        "b62b3b2cf2295dbbb24bf132d37fffc17a0bdccc84bbea6ea34737a236b58248"
        "a7ab4c14363c20541cb45338677f339782b205e355189658dd45ea3e8005f851"
        "148e7e15f431904fa79c6827ee426d3ab9cba936eb33d485db88a6f0ff9722a6"
        "d62ff788347e27c82e9e139eb03ae521539bf3d363b4baea76e5e884cf66fe6b"
        "cd8a9e089d36405db99d01db20464fb6cabbdce4f67e4ec3742f913a1dd2dac5"
    )
)


def _ensure_base_path(basePath: str) -> None:
    Expects(basePath is not None and basePath != "", "No folder provided")


class TDesktop(BaseObject):
    """Telegram Desktop client.

    Supports up to 3 accounts. Handles loading/saving tdata folders
    and converting between TelegramClient sessions.
    """

    kMaxAccounts: int = 3
    kDefaultKeyFile: str = "data"
    kPerformanceMode: bool = True

    def __init__(
        self,
        basePath: str = None,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        passcode: str = None,
        keyFile: str = None,
    ) -> None:
        self.__accounts: typing.List[td.Account] = []
        self.__basePath = basePath
        self.__keyFile = keyFile if (keyFile is not None) else TDesktop.kDefaultKeyFile
        self.__passcode = passcode if (passcode is not None) else str("")
        self.__passcodeBytes = self.__passcode.encode("ascii")
        self.__mainAccount: Optional[td.Account] = None
        self.__active_index = -1
        self.__passcodeKey = None
        self.__localKey = None
        self.__AppVersion = None
        self.__isLoaded = False
        self.__api = api.copy()

        if basePath is not None:
            self.__basePath = td.Storage.GetAbsolutePath(basePath)
            self.LoadTData()

    def isLoaded(self) -> bool:
        """Return True if accounts have been loaded from tdata or TelegramClient."""
        return self.__isLoaded

    def LoadTData(
        self, basePath: str = None, passcode: str = None, keyFile: str = None
    ):
        """Load accounts from a tdata folder.

        Raises:
            TDataBadDecryptKey: If the tdata is password-encrypted and
                passcode is missing or incorrect.
        """
        if basePath is None:
            basePath = self.basePath

        _ensure_base_path(basePath)

        if keyFile is not None and self.__keyFile != keyFile:
            self.__keyFile = keyFile

        if passcode is not None and self.__passcode != passcode:
            self.__passcode = passcode
            self.__passcodeBytes = passcode.encode("ascii")

        try:
            self.__loadFromTData()

        except OpenTeleException as e:
            if isinstance(e, TDataBadDecryptKey):
                if self.passcode == "":
                    raise TDataBadDecryptKey(
                        "The tdata folder is password-encrypted, please the set the argument 'passcode' to decrypt it"
                    )
                else:
                    raise TDataBadDecryptKey(
                        "Failed to decrypt tdata folder because of invalid passcode"
                    )
            else:
                raise e

        Expects(self.isLoaded(), "Failed to load? Something went seriously wrong")

    def SaveTData(
        self, basePath: str = None, passcode: str = None, keyFile: str = None
    ) -> bool:
        """Save the client session to a tdata folder."""
        if basePath is None:
            basePath = self.basePath

        self.__keyFile = (
            keyFile
            if (keyFile is not None and self.keyFile != keyFile)
            else self.keyFile
        )

        _ensure_base_path(basePath)

        if passcode is not None and self.__passcode != passcode:
            self.__passcode = passcode
            self.__passcodeBytes = passcode.encode("ascii")
            self.__isLoaded = False

        if not self.isLoaded():
            self.__generateLocalKey()
        Expects(self.isLoaded(), "Failed to load? Something went seriously wrong")

        try:
            basePath = td.Storage.GetAbsolutePath(basePath)
            if not self.basePath:
                self.__basePath = basePath

            self.__writeAccounts(basePath, keyFile)
            return True

        except OpenTeleException as e:
            raise TDataSaveFailed("Could not save tdata, something went wrong") from e

    def __writeAccounts(self, basePath: str, keyFile: str = None) -> None:
        Expects(len(self.accounts) > 0)
        _ensure_base_path(basePath)

        for account in self.accounts:
            account._writeData(basePath, keyFile)

        key = td.Storage.FileWriteDescriptor("key_" + self.keyFile, basePath)
        key.writeData(self.__passcodeKeySalt)
        key.writeData(self.__passcodeKeyEncrypted)

        keySize = sizeof(int32) + sizeof(int32) * len(self.accounts)
        keyData = td.Storage.EncryptedDescriptor(keySize)
        keyData.stream.writeInt32(len(self.accounts))

        for account in self.accounts:
            keyData.stream.writeInt32(account.index)

        keyData.stream.writeInt32(self.__active_index)
        key.writeEncrypted(keyData, self.__localKey)  # type: ignore
        key.finish()

    def __generateLocalKey(self) -> None:
        Expects(not self.isLoaded())

        if self.kPerformanceMode and not self.__passcodeBytes:
            self.__localKey = td.AuthKey(_PERF_LOCAL_KEY)
            self.__passcodeKeySalt = QByteArray(_PERF_PASSCODE_KEY_SALT)
            self.__passcodeKey = td.AuthKey(_PERF_PASSCODE_KEY)
            self.__passcodeKeyEncrypted = QByteArray(_PERF_PASSCODE_KEY_ENCRYPTED)
        else:
            LocalEncryptSaltSize = 32

            _pass = td.Storage.RandomGenerate(td.AuthKey.kSize)
            _salt = td.Storage.RandomGenerate(LocalEncryptSaltSize)
            self.__localKey = td.Storage.CreateLocalKey(_salt, _pass)

            self.__passcodeKeySalt = td.Storage.RandomGenerate(LocalEncryptSaltSize)
            self.__passcodeKey = td.Storage.CreateLocalKey(
                self.__passcodeKeySalt, QByteArray(self.__passcodeBytes)
            )

            passKeyData = td.Storage.EncryptedDescriptor(td.AuthKey.kSize)
            self.__localKey.write(passKeyData.stream)

            self.__passcodeKeyEncrypted = td.Storage.PrepareEncrypted(
                passKeyData, self.__passcodeKey
            )

            for account in self.accounts:
                account.localKey = self.localKey

        self.__isLoaded = True

    def _addSingleAccount(self, account: td.Account):
        Expects(
            account.isLoaded(),
            "Could not add account because the account hasn't been loaded",
        )

        account.localKey = self.localKey

        self.__accounts.append(account)

        if self.mainAccount is None:
            self.__mainAccount = self.__accounts[0]

    def __loadFromTData(self) -> None:
        _ensure_base_path(self.basePath)

        self.accounts.clear()

        keyData = td.Storage.ReadFile("key_" + self.keyFile, self.basePath)  # type: ignore

        salt, keyEncrypted, infoEncrypted = QByteArray(), QByteArray(), QByteArray()

        keyData.stream >> salt >> keyEncrypted >> infoEncrypted

        Expects(
            keyData.stream.status() == QDataStream.Status.Ok,
            QDataStreamFailed("Failed to stream keyData"),
        )

        self.__AppVersion = keyData.version
        self.__passcodeKey = td.Storage.CreateLocalKey(
            salt, QByteArray(self.__passcodeBytes)
        )

        keyInnerData = td.Storage.DecryptLocal(keyEncrypted, self.passcodeKey)  # type: ignore

        self.__localKey = td.AuthKey(keyInnerData.stream.readRawData(256))
        self.__passcodeKeyEncrypted = keyEncrypted
        self.__passcodeKeySalt = salt

        info = td.Storage.DecryptLocal(infoEncrypted, self.localKey)  # type: ignore
        count = info.stream.readInt32()

        Expects(count > 0, "accountsCount is zero, the data might has been corrupted")

        for i in range(count):
            index = info.stream.readInt32()
            if (index >= 0) and (index < TDesktop.kMaxAccounts):
                try:
                    account = td.Account(
                        self,
                        basePath=self.basePath,
                        api=self.api,
                        keyFile=self.keyFile,
                        index=index,
                    )
                    account.prepareToStart(self.__localKey)

                    if account.isLoaded():
                        self.accounts.append(account)

                except OpenTeleException:
                    pass

        Expects(len(self.accounts) > 0, "No account has been loaded")

        self.__active_index = 0
        if not info.stream.atEnd():
            self.__active_index = info.stream.readInt32()

        for account in self.accounts:
            if account.index == self.__active_index:
                self.__mainAccount = account
                break

        if not self.__mainAccount:
            self.__mainAccount = self.accounts[0]

        self.__isLoaded = True

    @typing.overload
    async def ToTelethon(
        self,
        session: Union[str, Session] = None,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
        *,
        connection: typing.Type[Connection] = ConnectionTcpFull,
        use_ipv6: bool = False,
        proxy: Union[tuple, dict] = None,
        local_addr: Union[str, tuple] = None,
        timeout: int = 10,
        request_retries: int = 5,
        connection_retries: int = 5,
        retry_delay: int = 1,
        auto_reconnect: bool = True,
        sequential_updates: bool = False,
        flood_sleep_threshold: int = 60,
        raise_last_call_error: bool = False,
        loop: asyncio.AbstractEventLoop = None,
        base_logger: Union[str, logging.Logger] = None,
        receive_updates: bool = True,
    ) -> tl.TelegramClient: ...

    async def ToTelethon(  # noqa: F811
        self,
        session: Union[str, Session] = None,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
        **kwargs,
    ) -> tl.TelegramClient:
        """Convert this TDesktop to a TelegramClient."""
        Expects(
            self.isLoaded(),
            TDesktopNotLoaded("You need to load accounts from a tdata folder first"),
        )
        Expects(
            self.accountsCount > 0,
            TDesktopHasNoAccount("There is no account in this instance of TDesktop"),
        )
        assert self.mainAccount

        return await tl.TelegramClient.FromTDesktop(
            self.mainAccount,
            session=session,
            flag=flag,
            api=api,
            password=password,
            **kwargs,
        )

    @staticmethod
    async def FromTelethon(
        telethonClient: tl.TelegramClient,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
    ) -> TDesktop:
        """Create a TDesktop instance from a TelegramClient."""
        Expects(
            (flag == CreateNewSession) or (flag == UseCurrentSession),
            LoginFlagInvalid("LoginFlag invalid"),
        )

        _self = TDesktop()
        _self.__generateLocalKey()

        await td.Account.FromTelethon(
            telethonClient, flag=flag, api=api, password=password, owner=_self
        )

        return _self

    @staticmethod
    async def FromSessionJson(
        session_path: str,
        json_path: str = None,
        flag: Type[LoginFlag] = UseCurrentSession,
        api: Union[Type[APIData], APIData] = None,
        password: str = None,
    ) -> TDesktop:
        """Create a TDesktop instance from .session + .json files.

        Args:
            session_path: Path to the .session file.
            json_path: Path to the .json metadata file. If None, derived
                       from session_path.
            flag: UseCurrentSession or CreateNewSession.
            api: API to use for the TDesktop. If None, uses the API
                 from the JSON file.
            password: 2FA password for CreateNewSession.

        Returns:
            A loaded TDesktop instance.
        """
        client = await tl.TelegramClient.FromSessionJson(
            session_path, json_path, flag=UseCurrentSession, password=password
        )
        use_api = api if api is not None else (client._api_data or API.TelegramDesktop)
        return await TDesktop.FromTelethon(
            client, flag=flag, api=use_api, password=password
        )

    async def SaveSessionJson(
        self,
        session_path: str,
        api: Union[Type[APIData], APIData] = None,
        fetch_user_info: bool = False,
    ) -> typing.Tuple[str, str]:
        """Save this TDesktop's session to .session + .json files.

        Args:
            session_path: Output path for the .session file.
            api: APIData for the JSON metadata. Defaults to self.api.
            fetch_user_info: If True and connected, fetch user info.

        Returns:
            Tuple of (session_file_path, json_file_path).
        """
        Expects(
            self.isLoaded(),
            TDesktopNotLoaded("You need to load accounts from a tdata folder first"),
        )
        Expects(
            self.accountsCount > 0,
            TDesktopHasNoAccount("There is no account in this instance of TDesktop"),
        )

        use_api = api if api is not None else self.api
        client = await self.ToTelethon(flag=UseCurrentSession, api=use_api)
        return await client.SaveSessionJson(
            session_path, api=use_api, fetch_user_info=fetch_user_info
        )

    @classmethod
    def PerformanceMode(cls, enabled: bool = True):
        """Enable or disable performance mode.

        When enabled, uses a constant localKey instead of generating one each time,
        making SaveTData() ~5000x faster. Disabled automatically when passcode is set.
        """
        cls.kPerformanceMode = enabled

    @property
    def api(self) -> APIData:
        return self.__api

    @api.setter
    def api(self, value) -> None:
        self.__api = value
        for account in self.accounts:
            account.api = value

    @property
    def basePath(self) -> Optional[str]:
        return self.__basePath

    @property
    def passcode(self) -> str:
        return self.__passcode

    @property
    def keyFile(self) -> str:
        return self.__keyFile

    @property
    def passcodeKey(self) -> Optional[td.AuthKey]:
        return self.__passcodeKey

    @property
    def localKey(self) -> Optional[td.AuthKey]:
        return self.__localKey

    @property
    def AppVersion(self) -> Optional[int]:
        return self.__AppVersion

    @property
    def accountsCount(self) -> int:
        return len(self.__accounts)

    @property
    def accounts(self) -> List[td.Account]:
        return self.__accounts

    @property
    def mainAccount(self) -> Optional[td.Account]:
        return self.__mainAccount



from __future__ import annotations

from ..utils import BaseObject
from ..exception import (
    Expects,
    LoginFlagInvalid,
    MaxAccountLimit,
    QDataStreamFailed,
    TDAccountNotLoaded,
    TDataAuthKeyNotFound,
    TelethonUnauthorized,
)
from ..api import API, APIData, CreateNewSession, LoginFlag, UseCurrentSession
from .configs import DcId
from . import shared as td
from .. import tl
from .map_data import MapData
from .storage_account import StorageAccount
from typing import List, Optional, Type, Union
from ..qt_compat import QByteArray, QDataStream, QIODevice
import asyncio
import typing

from telethon.network.connection.connection import Connection
from telethon.network.connection.tcpfull import ConnectionTcpFull
from telethon.sessions.abstract import Session

import logging


class Account(BaseObject):
    """Telegram Desktop account."""

    kWideIdsTag: int = int(~0)

    def __init__(
        self,
        owner: td.TDesktop,
        basePath: str = None,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        keyFile: str = None,
        index: int = 0,
    ) -> None:
        self.__owner = owner
        self.__localKey = None
        self.__authKey = None
        self.__isLoaded = False
        self.__isAuthorized = False
        self.__UserId = 0
        self.__MainDcId = DcId(0)

        self.__basePath = td.Storage.GetAbsolutePath(basePath)
        self.__keyFile = (
            keyFile if (keyFile is not None) else td.TDesktop.kDefaultKeyFile
        )

        self.__mtpKeys: typing.List[td.AuthKey] = []
        self.__mtpKeysToDestroy: typing.List[td.AuthKey] = []
        self.api = api

        self._local = StorageAccount(
            self, self.basePath, td.Storage.ComposeDataString(self.__keyFile, index)
        )
        self.index = index

    @property
    def api(self) -> APIData:
        return self.__api

    @api.setter
    def api(self, value) -> None:
        self.__api = value
        if self.owner.api != self.api:
            self.owner.api = self.api

    @property
    def owner(self) -> td.TDesktop:
        return self.__owner

    @property
    def basePath(self) -> str:
        return self.__basePath

    @property
    def keyFile(self) -> str:
        return self.__keyFile

    @keyFile.setter
    def keyFile(self, value):
        self.__keyFile = value
        self._local.keyFile = td.Storage.ComposeDataString(value, self.index)

    @property
    def localKey(self) -> Optional[td.AuthKey]:
        return self.__localKey

    @localKey.setter
    def localKey(self, value):
        self.__localKey = value
        self._local.localKey = value

    @property
    def authKey(self) -> Optional[td.AuthKey]:
        return self.__authKey

    @property
    def UserId(self) -> int:
        return self.__UserId

    @property
    def MainDcId(self) -> DcId:
        return self.__MainDcId

    @property
    def MtpConfig(self) -> td.MTP.Config:
        return self._local.config

    @property
    def MapData(self) -> MapData:
        return self._local.mapData

    def isAuthorized(self) -> bool:
        return self.__isAuthorized

    def isLoaded(self) -> bool:
        return self.__isLoaded

    def prepareToStart(self, localKey: td.AuthKey) -> td.MTP.Config:
        self.__localKey = localKey
        return self._local.start(localKey)

    def _findMainAuthKey(self):
        for key in self.__mtpKeys:
            if key.dcId == self.MainDcId:
                self.__authKey = key
                return
        raise TDataAuthKeyNotFound("Could not find the main authKey")

    def _setMtpAuthorizationCustom(
        self,
        dcId: DcId,
        userId: int,
        mtpKeys: List[td.AuthKey],
        mtpKeysToDestroy: List[td.AuthKey] = None,
    ):
        if mtpKeysToDestroy is None:
            mtpKeysToDestroy = []
        self.__MainDcId = dcId
        self.__UserId = userId
        self.__mtpKeys = mtpKeys
        self._findMainAuthKey()
        self.__isLoaded = True

    def _setMtpAuthorization(self, serialized: QByteArray):
        stream = QDataStream(serialized)
        stream.setVersion(QDataStream.Version.Qt_5_1)

        self.__UserId = stream.readInt32()
        self.__MainDcId = DcId(stream.readInt32())

        if ((self.__UserId << 32) | self.__MainDcId) == Account.kWideIdsTag:
            self.__UserId = stream.readUInt64()
            self.__MainDcId = DcId(stream.readInt32())

        Expects(
            stream.status() == QDataStream.Status.Ok,
            QDataStreamFailed("Could not read main fields from mtp authorization."),
        )

        def readKeys(keys: typing.List[td.AuthKey]):
            key_count = stream.readInt32()
            Expects(
                stream.status() == QDataStream.Status.Ok,
                QDataStreamFailed("Could not read keys count from mtp authorization."),
            )
            for i in range(key_count):
                dcId = DcId(stream.readInt32())
                keys.append(
                    td.AuthKey.FromStream(stream, td.AuthKeyType.ReadFromFile, dcId)
                )

        self.__mtpKeys.clear()
        self.__mtpKeysToDestroy.clear()

        readKeys(self.__mtpKeys)
        readKeys(self.__mtpKeysToDestroy)

        self._findMainAuthKey()
        self.__isLoaded = True

    def serializeMtpAuthorization(self) -> QByteArray:
        Expects(self.isLoaded(), "Account data not loaded yet")

        def writeKeys(stream: QDataStream, keys: typing.List[td.AuthKey]):
            stream.writeInt32(len(keys))
            for key in keys:
                stream.writeInt32(key.dcId)
                stream.writeRawData(key.key)

        result = QByteArray()
        stream = QDataStream(result, QIODevice.OpenModeFlag.WriteOnly)
        stream.setVersion(QDataStream.Version.Qt_5_1)

        stream.writeInt64(Account.kWideIdsTag)
        stream.writeInt64(self.UserId)
        stream.writeInt32(self.MainDcId)

        writeKeys(stream, self.__mtpKeys)
        writeKeys(stream, self.__mtpKeysToDestroy)
        return result

    def _writeData(self, baseGlobalPath: str, keyFile: str = None) -> None:
        self._local._writeData(baseGlobalPath, keyFile)

    def SaveTData(
        self, basePath: str = None, passcode: str = None, keyFile: str = None
    ) -> None:
        """Save this account to a tdata folder."""
        if basePath is None:
            basePath = self.basePath

        basePath = td.Storage.GetAbsolutePath(basePath)

        if self.basePath is None:
            self.__basePath = basePath

        self.owner.SaveTData(basePath, passcode, keyFile)

    @typing.overload
    async def ToTelethon(
        self,
        session: Union[str, Session] = None,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
        *,
        connection: typing.Type[Connection] = ConnectionTcpFull,
        use_ipv6: bool = False,
        proxy: Union[tuple, dict] = None,
        local_addr: Union[str, tuple] = None,
        timeout: int = 10,
        request_retries: int = 5,
        connection_retries: int = 5,
        retry_delay: int = 1,
        auto_reconnect: bool = True,
        sequential_updates: bool = False,
        flood_sleep_threshold: int = 60,
        raise_last_call_error: bool = False,
        loop: asyncio.AbstractEventLoop = None,
        base_logger: Union[str, logging.Logger] = None,
        receive_updates: bool = True,
    ) -> tl.TelegramClient:
        pass

    async def ToTelethon(  # noqa: F811
        self,
        session: Union[str, Session] = None,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
        **kwargs,
    ) -> tl.TelegramClient:
        Expects(
            self.isLoaded(),
            TDAccountNotLoaded("Account not loaded yet"),
        )

        return await tl.TelegramClient.FromTDesktop(
            self, session=session, flag=flag, api=api, password=password, **kwargs
        )

    @staticmethod
    async def FromTelethon(
        telethonClient: tl.TelegramClient,
        flag: Type[LoginFlag] = CreateNewSession,
        api: Union[Type[APIData], APIData] = API.TelegramDesktop,
        password: str = None,
        owner: td.TDesktop = None,
    ):
        Expects(
            (flag == CreateNewSession) or (flag == UseCurrentSession),
            LoginFlagInvalid("LoginFlag invalid"),
        )

        if flag == CreateNewSession:
            if not telethonClient.is_connected():
                try:
                    await telethonClient.connect()
                except OSError as e:
                    raise TelethonUnauthorized(
                        "Could not connect telethon client to the server"
                    ) from e

                Expects(
                    await telethonClient.is_user_authorized(),
                    exception=TelethonUnauthorized("Telethon client is unauthorized"),
                )

            copy = await telethonClient.QRLoginToNewClient(api=api, password=password)
            await copy.get_me()
        else:
            copy = telethonClient

        ss = copy.session
        authKey = ss.auth_key.key
        dcId = DcId(ss.dc_id)
        userId = copy.UserId
        authKey = td.AuthKey(authKey, td.AuthKeyType.ReadFromFile, dcId)

        if userId is None:
            await copy.connect()
            await copy.get_me()
            userId = copy.UserId

        if owner is not None:
            Expects(
                owner.accountsCount < td.TDesktop.kMaxAccounts,
                exception=MaxAccountLimit(
                    "You can't have more than 3 accounts in one TDesktop client"
                ),
            )

            index = owner.accountsCount
            newAccount = Account(
                owner=owner,
                basePath=owner.basePath,
                api=api,
                keyFile=owner.keyFile,
                index=index,
            )
            newAccount._setMtpAuthorizationCustom(dcId, userId, [authKey])  # type: ignore
            owner._addSingleAccount(newAccount)
        else:
            index = 0
            newOwner = td.TDesktop()
            newAccount = Account(owner=newOwner, api=api, index=index)
            newAccount._setMtpAuthorizationCustom(dcId, userId, [authKey])  # type: ignore
            newOwner._addSingleAccount(newAccount)

        return newAccount



from __future__ import annotations

import logging
import typing
from ctypes import sizeof, c_uint32 as uint32, c_uint64 as uint64
from typing import Dict

from ..qt_compat import QByteArray

from ..exception import (
    Expects,
    ExpectStreamStatus,
    OpenTeleException,
    TDataReadMapDataFailed,
    TDataReadMapDataIncorrectPasscode,
)
from ..utils import BaseObject
from .configs import FileKey, PeerId, lskType
from . import shared as td


class MapData(BaseObject):  # nocov
    _HARDCODED_SETTINGS_KEY = FileKey(1851671142505648812)
    _LEGACY_SALT_SIZE = 32

    _KEY_TO_ATTR = {
        "locationsKey": "_locationsKey",
        "trustedBotsKey": "_trustedBotsKey",
        "recentStickersKeyOld": "_recentStickersKeyOld",
        "installedStickersKey": "_installedStickersKey",
        "featuredStickersKey": "_featuredStickersKey",
        "recentStickersKey": "_recentStickersKey",
        "favedStickersKey": "_favedStickersKey",
        "archivedStickersKey": "_archivedStickersKey",
        "savedGifsKey": "_savedGifsKey",
        "installedMasksKey": "_installedMasksKey",
        "recentMasksKey": "_recentMasksKey",
        "archivedMasksKey": "_archivedMasksKey",
        "installedCustomEmojiKey": "_installedCustomEmojiKey",
        "featuredCustomEmojiKey": "_featuredCustomEmojiKey",
        "archivedCustomEmojiKey": "_archivedCustomEmojiKey",
        "searchSuggestionsKey": "_searchSuggestionsKey",
        "webviewStorageTokenBots": "_webviewStorageTokenBots",
        "webviewStorageTokenOther": "_webviewStorageTokenOther",
        "legacyBackgroundKeyDay": "_legacyBackgroundKeyDay",
        "legacyBackgroundKeyNight": "_legacyBackgroundKeyNight",
        "userSettingsKey": "_settingsKey",
        "recentHashtagsAndBotsKey": "_recentHashtagsAndBotsKey",
        "exportSettingsKey": "_exportSettingsKey",
    }

    _LSK_KEY_MAP = {
        lskType.lskLocations: ("locationsKey",),
        lskType.lskTrustedBots: ("trustedBotsKey",),
        lskType.lskRecentStickersOld: ("recentStickersKeyOld",),
        lskType.lskBackgroundOldOld: ("legacyBackgroundKeyDay",),
        lskType.lskUserSettings: ("userSettingsKey",),
        lskType.lskRecentHashtagsAndBots: ("recentHashtagsAndBotsKey",),
        lskType.lskStickersOld: ("installedStickersKey",),
        lskType.lskFavedStickers: ("favedStickersKey",),
        lskType.lskSavedGifs: ("savedGifsKey",),
        lskType.lskExportSettings: ("exportSettingsKey",),
        lskType.lskSearchSuggestions: ("searchSuggestionsKey",),
        lskType.lskBackgroundOld: (
            "legacyBackgroundKeyDay",
            "legacyBackgroundKeyNight",
        ),
        lskType.lskStickersKeys: (
            "installedStickersKey",
            "featuredStickersKey",
            "recentStickersKey",
            "archivedStickersKey",
        ),
        lskType.lskMasksKeys: (
            "installedMasksKey",
            "recentMasksKey",
            "archivedMasksKey",
        ),
        lskType.lskCustomEmojiKeys: (
            "installedCustomEmojiKey",
            "featuredCustomEmojiKey",
            "archivedCustomEmojiKey",
        ),
    }

    _LSK_SKIP_TYPES = frozenset(
        {
            lskType.lskReportSpamStatusesOld,
            lskType.lskSavedGifsOld,
            lskType.lskSavedPeersOld,
        }
    )

    def __init__(self, basePath: str) -> None:
        self.basePath = basePath

        self._draftsMap: Dict[PeerId, FileKey] = {}
        self._draftCursorsMap: Dict[PeerId, FileKey] = {}
        self._draftsNotReadMap: Dict[PeerId, bool] = {}

        self._locationsKey = FileKey(0)
        self._trustedBotsKey = FileKey(0)
        self._installedStickersKey = FileKey(0)
        self._featuredStickersKey = FileKey(0)
        self._recentStickersKey = FileKey(0)
        self._favedStickersKey = FileKey(0)
        self._archivedStickersKey = FileKey(0)
        self._archivedMasksKey = FileKey(0)
        self._installedCustomEmojiKey = FileKey(0)
        self._featuredCustomEmojiKey = FileKey(0)
        self._archivedCustomEmojiKey = FileKey(0)
        self._searchSuggestionsKey = FileKey(0)
        self._webviewStorageTokenBots = FileKey(0)
        self._webviewStorageTokenOther = FileKey(0)
        self._savedGifsKey = FileKey(0)
        self._recentStickersKeyOld = FileKey(0)
        self._legacyBackgroundKeyDay = FileKey(0)
        self._legacyBackgroundKeyNight = FileKey(0)
        self._settingsKey = self._HARDCODED_SETTINGS_KEY
        self._recentHashtagsAndBotsKey = FileKey(0)
        self._exportSettingsKey = FileKey(0)
        self._installedMasksKey = FileKey(0)
        self._recentMasksKey = FileKey(0)

    def read(self, localKey: td.AuthKey, legacyPasscode: QByteArray) -> None:
        try:
            mapData = td.Storage.ReadFile("map", self.basePath)
        except OpenTeleException as e:
            raise TDataReadMapDataFailed(
                "Could not read map data, file not found or couldn't be opened"
            ) from e

        legacySalt, legacyKeyEncrypted, mapEncrypted = (
            QByteArray(),
            QByteArray(),
            QByteArray(),
        )

        mapData.stream >> legacySalt >> legacyKeyEncrypted >> mapEncrypted
        ExpectStreamStatus(mapData.stream, "Could not stream data from mapData")

        if not localKey:
            Expects(
                legacySalt.size() == self._LEGACY_SALT_SIZE,
                TDataReadMapDataFailed(
                    f"Bad salt in map file, size: {legacySalt.size()}"
                ),
            )

            legacyPasscodeKey = td.Storage.CreateLegacyLocalKey(
                legacySalt, legacyPasscode
            )

            try:
                keyData = td.Storage.DecryptLocal(legacyKeyEncrypted, legacyPasscodeKey)
            except OpenTeleException as e:
                raise TDataReadMapDataIncorrectPasscode(
                    "Could not decrypt pass-protected key from map file, maybe bad password..."
                ) from e

            localKey = td.AuthKey.FromStream(keyData.stream)

        try:
            map = td.Storage.DecryptLocal(mapEncrypted, localKey)
        except OpenTeleException as e:
            raise TDataReadMapDataFailed("Could not decrypt map data") from e

        self._parseMapStream(map, localKey, mapData.version)

    def _parseMapStream(self, map, localKey, mapVersion) -> None:
        draftsMap: typing.Dict[PeerId, FileKey] = {}
        draftCursorsMap: typing.Dict[PeerId, FileKey] = {}
        draftsNotReadMap: typing.Dict[PeerId, bool] = {}

        keys = {k: 0 for k in self._KEY_TO_ATTR}

        while not map.stream.atEnd():
            keyType = map.stream.readUInt32()

            if not self._readMapEntry(
                keyType, map, draftsMap, draftCursorsMap, draftsNotReadMap, keys
            ):
                break

            ExpectStreamStatus(map.stream, "Could not stream data from mapData")

        self._applyParsedKeys(
            localKey, mapVersion, draftsMap, draftCursorsMap, draftsNotReadMap, keys
        )

    def _readMapEntry(
        self, keyType, map, draftsMap, draftCursorsMap, draftsNotReadMap, keys
    ) -> bool:
        if keyType == lskType.lskDraft:
            count = map.stream.readUInt32()
            for i in range(count):
                key = FileKey(map.stream.readUInt64())
                peerIdSerialized = map.stream.readUInt64()
                peerId = PeerId.FromSerialized(peerIdSerialized)
                draftsMap[peerId] = key
                draftsNotReadMap[peerId] = True

        elif keyType == lskType.lskSelfSerialized:
            selfSerialized = QByteArray()
            map.stream >> selfSerialized

        elif keyType == lskType.lskDraftPosition:
            count = map.stream.readUInt32()
            for i in range(count):
                key = FileKey(map.stream.readUInt64())
                peerIdSerialized = map.stream.readUInt64()
                peerId = PeerId.FromSerialized(peerIdSerialized)
                draftCursorsMap[peerId] = key

        elif keyType in (
            lskType.lskLegacyImages,
            lskType.lskLegacyStickerImages,
            lskType.lskLegacyAudios,
        ):
            count = map.stream.readUInt32()
            for i in range(count):
                map.stream.readUInt64()
                map.stream.readUInt64()
                map.stream.readUInt64()
                map.stream.readInt32()

        elif keyType in self._LSK_KEY_MAP:
            for key_name in self._LSK_KEY_MAP[keyType]:
                keys[key_name] = map.stream.readUInt64()

        elif keyType in self._LSK_SKIP_TYPES:
            map.stream.readUInt64()

        elif keyType == lskType.lskWebviewTokens:
            return False

        else:
            logging.warning(f"Unknown key type in encrypted map: {keyType}")

        return True

    def _applyParsedKeys(
        self, localKey, mapVersion, draftsMap, draftCursorsMap, draftsNotReadMap, keys
    ) -> None:
        self.__localKey = localKey
        self._draftsMap = draftsMap
        self._draftCursorsMap = draftCursorsMap
        self._draftsNotReadMap = draftsNotReadMap
        for key_name, attr_name in self._KEY_TO_ATTR.items():
            setattr(self, attr_name, keys[key_name])
        self._oldMapVersion = mapVersion

    def prepareToWrite(self) -> td.Storage.EncryptedDescriptor:
        mapSize = self._calculateMapSize()
        mapData = td.Storage.EncryptedDescriptor(mapSize)
        self._writeMapEntries(mapData.stream)
        return mapData

    def _calculateMapSize(self) -> int:
        mapSize = 0

        if len(self._draftsMap) > 0:
            mapSize += sizeof(uint32) * 2 + len(self._draftsMap) * sizeof(uint64) * 2
        if len(self._draftCursorsMap) > 0:
            mapSize += (
                sizeof(uint32) * 2 + len(self._draftCursorsMap) * sizeof(uint64) * 2
            )

        single_key_fields = [
            self._locationsKey,
            self._trustedBotsKey,
            self._recentStickersKeyOld,
            self._favedStickersKey,
            self._savedGifsKey,
            self._settingsKey,
            self._recentHashtagsAndBotsKey,
            self._exportSettingsKey,
            self._searchSuggestionsKey,
        ]
        for key in single_key_fields:
            if key:
                mapSize += sizeof(uint32) + sizeof(uint64)

        mapSize += self._groupKeySize(
            4,
            self._installedStickersKey,
            self._featuredStickersKey,
            self._recentStickersKey,
            self._archivedStickersKey,
        )
        mapSize += self._groupKeySize(
            3, self._installedMasksKey, self._recentMasksKey, self._archivedMasksKey
        )
        mapSize += self._groupKeySize(
            3,
            self._installedCustomEmojiKey,
            self._featuredCustomEmojiKey,
            self._archivedCustomEmojiKey,
        )
        mapSize += self._groupKeySize(
            2, self._webviewStorageTokenBots, self._webviewStorageTokenOther
        )

        return mapSize

    def _writeMapEntries(self, stream) -> None:
        if len(self._draftsMap) > 0:
            stream.writeUInt32(lskType.lskDraft)
            stream.writeUInt32(len(self._draftsMap))
            for key, value in self._draftsMap.items():
                stream.writeUInt64(value)
                stream.writeUInt64(PeerId(key).Serialize())

        if len(self._draftCursorsMap) > 0:
            stream.writeUInt32(lskType.lskDraftPosition)
            stream.writeUInt32(len(self._draftCursorsMap))
            for key, value in self._draftCursorsMap.items():
                stream.writeUInt64(value)
                stream.writeUInt64(PeerId(key).Serialize())

        self._writeKeyIfSet(stream, lskType.lskLocations, self._locationsKey)
        self._writeKeyIfSet(stream, lskType.lskTrustedBots, self._trustedBotsKey)
        self._writeKeyIfSet(
            stream, lskType.lskRecentStickersOld, self._recentStickersKeyOld
        )

        self._writeKeyGroupIfAnySet(
            stream,
            lskType.lskStickersKeys,
            self._installedStickersKey,
            self._featuredStickersKey,
            self._recentStickersKey,
            self._archivedStickersKey,
        )

        self._writeKeyIfSet(stream, lskType.lskFavedStickers, self._favedStickersKey)
        self._writeKeyIfSet(stream, lskType.lskSavedGifs, self._savedGifsKey)
        self._writeKeyIfSet(stream, lskType.lskUserSettings, self._settingsKey)
        self._writeKeyIfSet(
            stream, lskType.lskRecentHashtagsAndBots, self._recentHashtagsAndBotsKey
        )
        self._writeKeyIfSet(stream, lskType.lskExportSettings, self._exportSettingsKey)

        self._writeKeyGroupIfAnySet(
            stream,
            lskType.lskMasksKeys,
            self._installedMasksKey,
            self._recentMasksKey,
            self._archivedMasksKey,
        )
        self._writeKeyGroupIfAnySet(
            stream,
            lskType.lskCustomEmojiKeys,
            self._installedCustomEmojiKey,
            self._featuredCustomEmojiKey,
            self._archivedCustomEmojiKey,
        )

        self._writeKeyIfSet(
            stream, lskType.lskSearchSuggestions, self._searchSuggestionsKey
        )

        self._writeKeyGroupIfAnySet(
            stream,
            lskType.lskWebviewTokens,
            self._webviewStorageTokenBots,
            self._webviewStorageTokenOther,
        )

    @staticmethod
    def _writeKeyIfSet(stream, keyType: int, keyValue) -> None:
        if keyValue:
            stream.writeUInt32(keyType)
            stream.writeUInt64(keyValue)

    @staticmethod
    def _writeKeyGroupIfAnySet(stream, keyType: int, *keys) -> None:
        if any(keys):
            stream.writeUInt32(keyType)
            for key in keys:
                stream.writeUInt64(key)

    @staticmethod
    def _groupKeySize(count: int, *keys) -> int:
        if any(keys):
            return sizeof(uint32) + count * sizeof(uint64)
        return 0



from ..exception import *  # noqa: F403
from ..utils import *  # noqa: F403
from ..api import (
    APIData as APIData,
    API as API,
    LoginFlag as LoginFlag,
    CreateNewSession as CreateNewSession,
    UseCurrentSession as UseCurrentSession,
)
from .. import td as td

from typing import (
    Union as Union,
    Callable as Callable,
    TypeVar as TypeVar,
    Type as Type,
    List as List,
    Dict as Dict,
    Any as Any,
    TYPE_CHECKING as TYPE_CHECKING,
)
from ctypes import sizeof as sizeof

import telethon as telethon
from telethon.sessions import StringSession as StringSession
from telethon.crypto import AuthKey as AuthKey
from telethon import tl as tl, functions as functions, types as types, utils as utils

from telethon.network.connection.connection import Connection as Connection
from telethon.network.connection.tcpfull import ConnectionTcpFull as ConnectionTcpFull

from telethon.sessions.abstract import Session as Session
from telethon.sessions.sqlite import SQLiteSession as SQLiteSession
from telethon.sessions.memory import MemorySession as MemorySession

import asyncio as asyncio



"""Fingerprint configuration and consistency module for opentele2.

Centralizes TL layer tracking, per-platform app versioning, strict mode,
initConnection parameter validation, and transport recommendations.

The constants here should be updated whenever Telegram releases major
client updates (typically every 1-3 months).
"""

from __future__ import annotations

import random
import time
import warnings
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple

__all__ = [
    "LAYER",
    "PlatformVersions",
    "StrictMode",
    "FingerprintConfig",
    "TransportRecommendation",
    "validate_init_connection_params",
    "get_recommended_layer",
    "get_platform_versions",
]


# ---------------------------------------------------------------------------
# Current TL Layer — update this with each major Telegram release
# ---------------------------------------------------------------------------
# Official schema layer as of February 2026 (from core.telegram.org/schema)
LAYER: int = 214

# Telethon may be slightly ahead or behind; we track both so the library
# can warn when they diverge significantly.
_TELETHON_LAYER: Optional[int] = None


def _detect_telethon_layer() -> int:
    """Detect the TL layer used by the installed Telethon version."""
    global _TELETHON_LAYER
    if _TELETHON_LAYER is None:
        try:
            from telethon.tl.alltlobjects import LAYER as _tl_layer

            _TELETHON_LAYER = _tl_layer
        except ImportError:
            _TELETHON_LAYER = LAYER
    return _TELETHON_LAYER


def get_recommended_layer() -> int:
    """Return the layer that should be used in ``invokeWithLayer``.

    Prefers the Telethon built-in layer (since it must match the generated
    TL objects) but warns if it is too far from the official schema.
    """
    telethon_layer = _detect_telethon_layer()
    diff = abs(telethon_layer - LAYER)
    if diff > 15:
        warnings.warn(
            f"Telethon TL layer ({telethon_layer}) differs from the latest "
            f"official schema layer ({LAYER}) by {diff} layers. "
            f"This may cause initConnection fingerprint inconsistencies. "
            f"Consider updating Telethon or opentele2.",
            stacklevel=2,
        )
    return telethon_layer


# ---------------------------------------------------------------------------
# Per-platform version constants — updated February 2026
# ---------------------------------------------------------------------------
# These reflect the *latest stable* versions of official Telegram clients.
# When a new release is pushed to Google Play / App Store / GitHub, update
# the matching entry here.


@dataclass(frozen=True)
class PlatformVersions:
    """Tracks the latest known official client versions per platform."""

    # Telegram for Android (Google Play / GitHub)
    android_app_version: str = "12.3.0"
    android_app_version_code: int = 57428  # build code sent in some requests
    android_sdk_range: Tuple[int, int] = (29, 35)  # API 29 (Android 10) – 35 (15)
    android_latest_sdk: int = 35

    # Telegram for iOS (App Store)
    ios_app_version: str = "12.3"
    ios_build_number: int = 35198
    ios_version_range: Tuple[str, str] = ("16.0", "26.2")

    # Telegram Desktop (GitHub releases)
    desktop_app_version: str = "5.12.3"
    desktop_app_version_suffix: str = "x64"  # or empty for 32-bit

    # Telegram macOS (Swift, App Store)
    macos_app_version: str = "12.3"
    macos_build_number: int = 256120

    # TelegramX for Android
    android_x_app_version: str = "12.3.0"

    # Web clients
    web_z_version: str = "5.0.0 Z"
    web_a_version: str = "5.0.0 A"
    web_k_version: str = "1.4.2 K"

    # Chrome user-agent for web clients (keep updated)
    chrome_version: str = "144.0.0.0"
    user_agent: str = (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
        "AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/144.0.0.0 Safari/537.36"
    )


# Singleton — import ``PLATFORM_VERSIONS`` for the latest known values.
PLATFORM_VERSIONS = PlatformVersions()


def get_platform_versions() -> PlatformVersions:
    """Return the current platform version constants."""
    return PLATFORM_VERSIONS


# ---------------------------------------------------------------------------
# initConnection field-order specification (per official client source)
# ---------------------------------------------------------------------------
# The TL constructor ``initConnection#c1cd5ea9`` has a strict binary field
# order. Telethon handles this correctly, but we document it here for
# auditing and strict-mode validation.
#
# Field order (flags-based):
#   flags:#  api_id:int  device_model:string  system_version:string
#   app_version:string  system_lang_code:string  lang_pack:string
#   lang_code:string  proxy:flags.0?InputClientProxy
#   params:flags.1?JSONValue  query:!X

_INIT_CONNECTION_FIELD_ORDER = [
    "api_id",
    "device_model",
    "system_version",
    "app_version",
    "system_lang_code",
    "lang_pack",
    "lang_code",
    "proxy",
    "params",
    "query",
]

# Valid lang_pack values for official clients
_VALID_LANG_PACKS = frozenset(
    {
        "android",
        "ios",
        "tdesktop",
        "macos",
        "",  # Web clients use empty string
    }
)

# Map: lang_pack -> expected api_id
_LANG_PACK_API_ID_MAP: Dict[str, int] = {
    "tdesktop": 2040,
    "android": 6,
    "ios": 10840,
    "macos": 2834,
}


def validate_init_connection_params(
    api_id: int,
    device_model: str,
    system_version: str,
    app_version: str,
    system_lang_code: str,
    lang_pack: str,
    lang_code: str,
    *,
    strict: bool = False,
) -> List[str]:
    """Validate initConnection parameters for consistency.

    Returns a list of warning strings. Empty list means all checks pass.
    If *strict* is True, checks are more aggressive (e.g. lang_pack must
    match api_id).
    """
    issues: List[str] = []

    # 1. lang_pack must be a known value
    if lang_pack not in _VALID_LANG_PACKS:
        issues.append(
            f"lang_pack '{lang_pack}' is not a known official value. "
            f"Expected one of: {sorted(_VALID_LANG_PACKS)}"
        )

    # 2. lang_code should be a valid IETF language tag (basic check)
    if not lang_code or len(lang_code) < 2:
        issues.append(f"lang_code '{lang_code}' looks invalid (too short)")

    # 3. system_lang_code should include region for mobile clients
    if lang_pack in ("android", "ios") and "-" not in system_lang_code:
        issues.append(
            f"system_lang_code '{system_lang_code}' should include a region "
            f"code (e.g. 'en-US') for {lang_pack} clients"
        )

    # 4. Empty strings are suspicious
    if not device_model:
        issues.append("device_model is empty")
    if not system_version:
        issues.append("system_version is empty")
    if not app_version:
        issues.append("app_version is empty")

    # 5. Strict: lang_pack <-> api_id consistency
    if strict and lang_pack in _LANG_PACK_API_ID_MAP:
        expected_api_id = _LANG_PACK_API_ID_MAP[lang_pack]
        if api_id != expected_api_id:
            issues.append(
                f"api_id {api_id} does not match lang_pack '{lang_pack}' "
                f"(expected {expected_api_id})"
            )

    # 6. Strict: app_version should match known current versions
    if strict:
        _check_version_consistency(
            lang_pack, app_version, system_version, device_model, issues
        )

    return issues


def _check_version_consistency(
    lang_pack: str,
    app_version: str,
    system_version: str,
    device_model: str,
    issues: List[str],
) -> None:
    """Deep consistency check for strict mode."""
    pv = PLATFORM_VERSIONS

    if lang_pack == "android":
        if not app_version.startswith(pv.android_app_version.split(".")[0]):
            issues.append(
                f"Android app_version '{app_version}' major does not match "
                f"latest known '{pv.android_app_version}'"
            )
        if "SDK" in system_version:
            try:
                sdk_num = int(system_version.replace("SDK ", ""))
                lo, hi = pv.android_sdk_range
                if not (lo <= sdk_num <= hi):
                    issues.append(
                        f"Android SDK {sdk_num} is outside the expected "
                        f"range [{lo}, {hi}]"
                    )
            except ValueError:
                pass

    elif lang_pack == "ios":
        if not app_version.startswith(pv.ios_app_version.split(".")[0]):
            issues.append(
                f"iOS app_version '{app_version}' major does not match "
                f"latest known '{pv.ios_app_version}'"
            )

    elif lang_pack == "tdesktop":
        if not app_version.startswith(pv.desktop_app_version.split(".")[0]):
            issues.append(
                f"Desktop app_version '{app_version}' major does not match "
                f"latest known '{pv.desktop_app_version}'"
            )

    elif lang_pack == "macos":
        if not app_version.startswith(pv.macos_app_version.split(".")[0]):
            issues.append(
                f"macOS app_version '{app_version}' major does not match "
                f"latest known '{pv.macos_app_version}'"
            )


# ---------------------------------------------------------------------------
# Strict Mode
# ---------------------------------------------------------------------------


class StrictMode(Enum):
    """Controls how aggressively opentele2 enforces fingerprint consistency."""

    OFF = "off"
    """No extra checks — Telethon defaults."""

    WARN = "warn"
    """Emit warnings for inconsistencies but do not block."""

    STRICT = "strict"
    """Raise exceptions for any inconsistency that could be detected
    server-side."""


# Global configuration
@dataclass
class FingerprintConfig:
    """Global fingerprint configuration for the opentele2 session.

    Instantiate once and pass it to ``TelegramClient`` or use the module-level
    default ``DEFAULT_CONFIG``.
    """

    strict_mode: StrictMode = StrictMode.WARN
    """How to handle consistency issues."""

    auto_validate: bool = True
    """Automatically validate initConnection parameters on connect."""

    preferred_transport: str = "obfuscated"
    """Preferred transport type.  Official clients in 2025-2026 use
    ''obfuscated'' (intermediate padded with obfuscation) by default.
    Telethon supports ``ConnectionTcpObfuscated``.
    
    Valid values: 'full', 'intermediate', 'abridged', 'obfuscated'.
    """

    warn_on_layer_mismatch: bool = True
    """Warn when Telethon layer differs from official by >5 layers."""

    layer_override: Optional[int] = None
    """If set, override the layer used in ``invokeWithLayer``.
    Use with caution — mismatched layers cause deserialization errors."""

    randomize_msg_id_offset: bool = True
    """Add a small random offset to msg_id generation to avoid
    predictability.  Official clients do this."""

    _msg_id_offset: int = field(default_factory=lambda: random.randint(0, 0xFFFF))

    def get_effective_layer(self) -> int:
        """Return the layer to use, respecting overrides."""
        if self.layer_override is not None:
            return self.layer_override
        return get_recommended_layer()

    def validate_params(self, **kwargs) -> None:
        """Validate initConnection parameters based on strict_mode.

        Keyword args should match ``initConnection`` fields:
        api_id, device_model, system_version, app_version,
        system_lang_code, lang_pack, lang_code.
        """
        if not self.auto_validate:
            return

        issues = validate_init_connection_params(
            strict=(self.strict_mode == StrictMode.STRICT),
            **kwargs,
        )

        if not issues:
            return

        msg = "initConnection fingerprint issues:\n" + "\n".join(
            f"  - {i}" for i in issues
        )

        if self.strict_mode == StrictMode.STRICT:
            raise ValueError(msg)
        elif self.strict_mode == StrictMode.WARN:
            warnings.warn(msg, stacklevel=3)


# Module-level default config
DEFAULT_CONFIG = FingerprintConfig()


# ---------------------------------------------------------------------------
# Transport Recommendations
# ---------------------------------------------------------------------------


class TransportRecommendation:
    """Provides transport/connection type recommendations for 2025-2026.

    Official mobile clients (Android, iOS) use **obfuscated intermediate**
    transport.  Desktop uses **obfuscated abridged** or **obfuscated
    intermediate** depending on version.

    Using plain ``ConnectionTcpFull`` or ``ConnectionTcpAbridged`` without
    obfuscation is increasingly risky for mass operations.
    """

    @staticmethod
    def get_connection_class(lang_pack: str = "tdesktop"):
        """Return the recommended Telethon Connection class for a platform.

        Falls back to ``ConnectionTcpObfuscated`` for all platforms as it
        provides obfuscation which is what official clients use.
        """
        try:
            from telethon.network.connection.tcpobfuscated import (
                ConnectionTcpObfuscated,
            )

            return ConnectionTcpObfuscated
        except ImportError:
            from telethon.network.connection.tcpfull import ConnectionTcpFull

            return ConnectionTcpFull

    @staticmethod
    def get_available_transports() -> Dict[str, Any]:
        """List all available Telethon transport classes."""
        transports = {}
        try:
            from telethon.network.connection.tcpfull import ConnectionTcpFull

            transports["full"] = ConnectionTcpFull
        except ImportError:
            pass
        try:
            from telethon.network.connection.tcpabridged import (
                ConnectionTcpAbridged,
            )

            transports["abridged"] = ConnectionTcpAbridged
        except ImportError:
            pass
        try:
            from telethon.network.connection.tcpintermediate import (
                ConnectionTcpIntermediate,
            )

            transports["intermediate"] = ConnectionTcpIntermediate
        except ImportError:
            pass
        try:
            from telethon.network.connection.tcpobfuscated import (
                ConnectionTcpObfuscated,
            )

            transports["obfuscated"] = ConnectionTcpObfuscated
        except ImportError:
            pass
        return transports


# ---------------------------------------------------------------------------
# msg_id helpers (for auditing, not direct patching of Telethon)
# ---------------------------------------------------------------------------


def generate_msg_id_offset() -> int:
    """Generate a random offset to add to msg_id base time.

    Official clients use ``time * 2^32`` plus a small counter.
    Adding a random offset within the valid range avoids trivially
    predictable msg_ids.
    """
    return random.randint(0, 0xFFFF)


def is_valid_msg_id(msg_id: int, *, from_client: bool = True) -> bool:
    """Check if a msg_id follows the official rules.

    Client msg_ids must be odd (``msg_id % 2 == 1``).
    Server msg_ids must be even (``msg_id % 2 == 0``).
    msg_id must be > 0 and roughly correspond to current unix time * 2^32.
    """
    if msg_id <= 0:
        return False
    parity_ok = (msg_id % 2 == 1) if from_client else (msg_id % 2 == 0)
    if not parity_ok:
        return False

    # Rough time check — msg_id encodes unix time in the upper 32 bits
    encoded_time = msg_id >> 32
    now = int(time.time())
    # Allow ±5 minutes drift
    return abs(encoded_time - now) < 300



from .telethon import TelegramClient as TelegramClient
from ..fingerprint import (
    LAYER as LAYER,
    FingerprintConfig as FingerprintConfig,
    StrictMode as StrictMode,
    DEFAULT_CONFIG as DEFAULT_CONFIG,
    PlatformVersions as PlatformVersions,
    PLATFORM_VERSIONS as PLATFORM_VERSIONS,
    TransportRecommendation as TransportRecommendation,
    validate_init_connection_params as validate_init_connection_params,
)
from ..consistency import (
    ConsistencyChecker as ConsistencyChecker,
    ConsistencyReport as ConsistencyReport,
    CheckResult as CheckResult,
)
